import os

import pytest
from django.urls import reverse

from ....test import assert_contains
from ...models import Setting

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
IMAGE = os.path.join(BASE_DIR, "testfiles", "image.png")
OTHER_IMAGE = os.path.join(BASE_DIR, "testfiles", "image-other.png")
OTHER_FILE = os.path.join(BASE_DIR, "testfiles", "other")


@pytest.fixture
def setting(db):
    obj = Setting.objects.get(setting="logo")
    yield obj
    obj.refresh_from_db()
    if obj.image:
        obj.image.delete(save=False)


def test_image_setting_can_be_set(admin_client, setting):
    with open(IMAGE, "rb") as image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {"forum_name": "Misago", "forum_address": "http://test.com", "logo": image},
        )

    setting.refresh_from_db()
    assert setting.image


def test_setting_image_also_sets_its_dimensions(admin_client, setting):
    with open(IMAGE, "rb") as image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {"forum_name": "Misago", "forum_address": "http://test.com", "logo": image},
        )

    setting.refresh_from_db()
    assert setting.image_width == 4
    assert setting.image_height == 2


def test_setting_image_filename_is_prefixed_with_setting_name(admin_client, setting):
    with open(IMAGE, "rb") as image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {"forum_name": "Misago", "forum_address": "http://test.com", "logo": image},
        )

    setting.refresh_from_db()
    assert ("%s." % setting.setting) in str(setting.image.name)


def test_setting_image_filename_is_hashed(admin_client, setting):
    with open(IMAGE, "rb") as image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {"forum_name": "Misago", "forum_address": "http://test.com", "logo": image},
        )

    setting.refresh_from_db()
    assert str(setting.image.name).endswith(".c68ed127.png")


def test_image_setting_rejects_non_image_file(admin_client, setting):
    with open(OTHER_FILE, "rb") as not_image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    setting.refresh_from_db()
    assert not setting.image


@pytest.fixture
def setting_with_value(admin_client, setting):
    with open(IMAGE, "rb") as not_image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    setting.refresh_from_db()
    return setting


def test_image_setting_value_is_rendered_in_form(admin_client, setting_with_value):
    response = admin_client.get(reverse("misago:admin:settings:general:index"))
    assert_contains(response, setting_with_value.image.url)


def test_invalid_file_is_not_set_as_value(admin_client, setting):
    with open(OTHER_FILE, "rb") as not_image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    setting.refresh_from_db()
    assert not setting.image


def test_uploading_invalid_file_doesnt_remove_already_set_image(
    admin_client, setting_with_value
):
    with open(OTHER_FILE, "rb") as not_image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    setting_with_value.refresh_from_db()
    assert setting_with_value.image


def test_set_image_is_still_rendered_when_invalid_file_was_uploaded(
    admin_client, setting_with_value
):
    with open(OTHER_FILE, "rb") as not_image:
        response = admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    assert_contains(response, setting_with_value.image.url)


def test_uploading_new_image_replaces_already_set_image(
    admin_client, setting_with_value
):
    with open(OTHER_IMAGE, "rb") as not_image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    setting_with_value.refresh_from_db()
    assert str(setting_with_value.image.name).endswith(".d18644ab.png")


def test_uploading_new_image_deletes_image_file(admin_client, setting_with_value):
    with open(OTHER_IMAGE, "rb") as not_image:
        admin_client.post(
            reverse("misago:admin:settings:general:index"),
            {
                "forum_name": "Misago",
                "forum_address": "http://test.com",
                "logo": not_image,
            },
        )

    assert not os.path.exists(setting_with_value.image.path)


def test_omitting_setting_value_doesnt_remove_already_set_image(
    admin_client, setting_with_value
):
    admin_client.post(
        reverse("misago:admin:settings:general:index"),
        {"forum_name": "Misago", "forum_address": "http://test.com"},
    )

    setting_with_value.refresh_from_db()
    assert setting_with_value.image


def test_set_image_can_be_deleted_by_extra_option(admin_client, setting_with_value):
    admin_client.post(
        reverse("misago:admin:settings:general:index"),
        {"forum_name": "Misago", "forum_address": "http://test.com", "logo_delete": 1},
    )

    setting_with_value.refresh_from_db()
    assert not setting_with_value.image


def test_using_image_deletion_option_deletes_image_file(
    admin_client, setting_with_value
):
    admin_client.post(
        reverse("misago:admin:settings:general:index"),
        {"forum_name": "Misago", "forum_address": "http://test.com", "logo_delete": 1},
    )

    assert not os.path.exists(setting_with_value.image.path)



from ..client import get_headers_dict


def test_empty_headers_dict_is_returned_for_none():
    headers = get_headers_dict(None)
    assert headers == {}


def test_empty_headers_dict_is_returned_for_empty_str():
    headers = get_headers_dict("")
    assert headers == {}


def test_empty_headers_dict_is_returned_for_empty_multiline_str():
    headers = get_headers_dict("  \n   \n  ")
    assert headers == {}


def test_header_is_returned_from_str():
    headers = get_headers_dict("Lorem: ipsum")
    assert headers == {"Lorem": "ipsum"}


def test_headers_str_content_is_cleaned():
    headers = get_headers_dict("  Lorem:   ips:um\n\n")
    assert headers == {"Lorem": "ips:um"}


def test_multiple_headers_are_returned_from_multiline_str():
    headers = get_headers_dict("Lorem: ipsum\nDolor: met")
    assert headers == {"Lorem": "ipsum", "Dolor": "met"}



from django import forms
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.utils.translation import gettext_lazy as _

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...acl.models import Role
from ...admin.forms import YesNoSwitch
from ...categories import PRIVATE_THREADS_ROOT_NAME
from ...categories.models import Category
from ..models import Thread

__all__ = [
    "allow_use_private_threads",
    "can_use_private_threads",
    "allow_see_private_thread",
    "can_see_private_thread",
    "allow_change_owner",
    "can_change_owner",
    "allow_add_participants",
    "can_add_participants",
    "allow_remove_participant",
    "can_remove_participant",
    "allow_add_participant",
    "can_add_participant",
    "allow_message_user",
    "can_message_user",
]


class PermissionsForm(forms.Form):
    legend = _("Private threads")

    can_use_private_threads = YesNoSwitch(label=_("Can use private threads"))
    can_start_private_threads = YesNoSwitch(label=_("Can start private threads"))
    max_private_thread_participants = forms.IntegerField(
        label=_("Max number of users invited to private thread"),
        help_text=_("Enter 0 to don't limit number of participants."),
        initial=3,
        min_value=0,
    )
    can_add_everyone_to_private_threads = YesNoSwitch(
        label=_("Can add everyone to threads"),
        help_text=_(
            "Allows user to add users that are blocking them to private threads."
        ),
    )
    can_report_private_threads = YesNoSwitch(
        label=_("Can report private threads"),
        help_text=_(
            "Allows user to report private threads they are "
            "participating in, making them accessible to moderators."
        ),
    )
    can_moderate_private_threads = YesNoSwitch(
        label=_("Can moderate private threads"),
        help_text=_(
            "Allows user to read, reply, edit and delete content "
            "in reported private threads."
        ),
    )


def change_permissions_form(role):
    if isinstance(role, Role) and role.special_role != "anonymous":
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {
        "can_use_private_threads": 0,
        "can_start_private_threads": 0,
        "max_private_thread_participants": 3,
        "can_add_everyone_to_private_threads": 0,
        "can_report_private_threads": 0,
        "can_moderate_private_threads": 0,
    }

    new_acl.update(acl)

    algebra.sum_acls(
        new_acl,
        roles=roles,
        key=key_name,
        can_use_private_threads=algebra.greater,
        can_start_private_threads=algebra.greater,
        max_private_thread_participants=algebra.greater_or_zero,
        can_add_everyone_to_private_threads=algebra.greater,
        can_report_private_threads=algebra.greater,
        can_moderate_private_threads=algebra.greater,
    )

    if not new_acl["can_use_private_threads"]:
        new_acl["can_start_private_threads"] = 0
        return new_acl

    private_category = Category.objects.private_threads()

    new_acl["visible_categories"].append(private_category.pk)
    new_acl["browseable_categories"].append(private_category.pk)

    if new_acl["can_moderate_private_threads"]:
        new_acl["can_see_reports"].append(private_category.pk)

    category_acl = {
        "can_see": 1,
        "can_browse": 1,
        "can_see_all_threads": 1,
        "can_see_own_threads": 0,
        "can_start_threads": new_acl["can_start_private_threads"],
        "can_reply_threads": 1,
        "can_edit_threads": 1,
        "can_edit_posts": 1,
        "can_hide_own_threads": 0,
        "can_hide_own_posts": 1,
        "thread_edit_time": 0,
        "post_edit_time": 0,
        "can_hide_threads": 0,
        "can_hide_posts": 0,
        "can_protect_posts": 0,
        "can_move_posts": 0,
        "can_merge_posts": 0,
        "can_pin_threads": 0,
        "can_close_threads": 0,
        "can_move_threads": 0,
        "can_merge_threads": 0,
        "can_approve_content": 0,
        "can_report_content": new_acl["can_report_private_threads"],
        "can_see_reports": 0,
        "can_see_posts_likes": 0,
        "can_like_posts": 0,
        "can_hide_events": 0,
    }

    if new_acl["can_moderate_private_threads"]:
        category_acl.update(
            {
                "can_edit_threads": 2,
                "can_edit_posts": 2,
                "can_hide_threads": 2,
                "can_hide_posts": 2,
                "can_protect_posts": 1,
                "can_merge_posts": 1,
                "can_see_reports": 1,
                "can_close_threads": 1,
                "can_hide_events": 2,
            }
        )

    new_acl["categories"][private_category.pk] = category_acl

    return new_acl


def add_acl_to_thread(user_acl, thread):
    if thread.thread_type.root_name != PRIVATE_THREADS_ROOT_NAME:
        return

    if not hasattr(thread, "participant"):
        thread.participants_list = []
        thread.participant = None

    thread.acl.update(
        {
            "can_start_poll": False,
            "can_change_owner": can_change_owner(user_acl, thread),
            "can_add_participants": can_add_participants(user_acl, thread),
        }
    )


def register_with(registry):
    registry.acl_annotator(Thread, add_acl_to_thread)


def allow_use_private_threads(user_acl):
    if user_acl["is_anonymous"]:
        raise PermissionDenied(_("You have to sign in to use private threads."))
    if not user_acl["can_use_private_threads"]:
        raise PermissionDenied(_("You can't use private threads."))


can_use_private_threads = return_boolean(allow_use_private_threads)


def allow_see_private_thread(user_acl, target):
    if user_acl["can_moderate_private_threads"]:
        can_see_reported = target.has_reported_posts
    else:
        can_see_reported = False

    can_see_participating = user_acl["user_id"] in [
        p.user_id for p in target.participants_list
    ]

    if not (can_see_participating or can_see_reported):
        raise Http404()


can_see_private_thread = return_boolean(allow_see_private_thread)


def allow_change_owner(user_acl, target):
    is_moderator = user_acl["can_moderate_private_threads"]
    is_owner = target.participant and target.participant.is_owner

    if not (is_owner or is_moderator):
        raise PermissionDenied(
            _("Only thread owner and moderators can change threads owners.")
        )

    if not is_moderator and target.is_closed:
        raise PermissionDenied(_("Only moderators can change closed threads owners."))


can_change_owner = return_boolean(allow_change_owner)


def allow_add_participants(user_acl, target):
    is_moderator = user_acl["can_moderate_private_threads"]

    if not is_moderator:
        if not target.participant or not target.participant.is_owner:
            raise PermissionDenied(
                _("You have to be thread owner to add new participants to it.")
            )

        if target.is_closed:
            raise PermissionDenied(
                _("Only moderators can add participants to closed threads.")
            )

    max_participants = user_acl["max_private_thread_participants"]
    current_participants = len(target.participants_list) - 1

    if current_participants >= max_participants:
        raise PermissionDenied(_("You can't add any more new users to this thread."))


can_add_participants = return_boolean(allow_add_participants)


def allow_remove_participant(user_acl, thread, target):
    if user_acl["can_moderate_private_threads"]:
        return

    if user_acl["user_id"] == target.id:
        return  # we can always remove ourselves

    if thread.is_closed:
        raise PermissionDenied(
            _("Only moderators can remove participants from closed threads.")
        )

    if not thread.participant or not thread.participant.is_owner:
        raise PermissionDenied(
            _("You have to be thread owner to remove participants from it.")
        )


can_remove_participant = return_boolean(allow_remove_participant)


def allow_add_participant(user_acl, target, target_acl):
    message_format = {"user": target.username}

    if not can_use_private_threads(target_acl):
        raise PermissionDenied(
            _("%(user)s can't participate in private threads.") % message_format
        )

    if user_acl["can_add_everyone_to_private_threads"]:
        return

    if user_acl["can_be_blocked"] and target.is_blocking(user_acl["user_id"]):
        raise PermissionDenied(_("%(user)s is blocking you.") % message_format)

    if target.can_be_messaged_by_nobody:
        raise PermissionDenied(
            _("%(user)s is not allowing invitations to private threads.")
            % message_format
        )

    if target.can_be_messaged_by_followed and not target.is_following(
        user_acl["user_id"]
    ):
        message = _("%(user)s limits invitations to private threads to followed users.")
        raise PermissionDenied(message % message_format)


can_add_participant = return_boolean(allow_add_participant)


def allow_message_user(user_acl, target, target_acl):
    allow_use_private_threads(user_acl)
    allow_add_participant(user_acl, target, target_acl)


can_message_user = return_boolean(allow_message_user)



import re
from typing import Union

from django.http import Http404
from django.urls import resolve

from .htmlparser import ElementNode, RootNode, TextNode

MISAGO_ATTACHMENT_VIEWS = ("misago:attachment", "misago:attachment-thumbnail")
URL_RE = re.compile(
    r"(https?://)?"
    r"(www\.)?"
    r"(\w+((-|_)\w+)?\.)?"
    r"\w+((_|-|\w)+)?(\.[a-z][a-z]+)"
    r"(:[1-9][0-9]+)?"
    r"([^\s<>\[\]\(\);:]+)?"
)


def linkify_texts(node: Union[RootNode, ElementNode]):
    # Skip link replacement in some nodes
    if node.tag in ("pre", "code", "a"):
        return

    new_children = []
    for child in node.children:
        if isinstance(child, TextNode):
            if URL_RE.search(child.text):
                new_children += replace_links_in_text(child.text)
            else:
                new_children.append(child)
        else:
            new_children.append(child)
            linkify_texts(child)

    node.children = new_children


def replace_links_in_text(text: str) -> list:
    nodes = []

    while True:
        match = URL_RE.search(text)
        if not match:
            if text:
                nodes.append(TextNode(text=text))
            return nodes

        start, end = match.span()
        url = text[start:end]

        # Append text between 0 and start to nodes
        if start > 0:
            nodes.append(TextNode(text=text[:start]))

        nodes.append(
            ElementNode(
                tag="a",
                attrs={"href": url},
                children=[
                    TextNode(text=strip_link_protocol(url)),
                ],
            )
        )

        text = text[end:]


def clean_links(
    request,
    result,
    node: Union[RootNode, ElementNode, TextNode],
    force_shva=False,
):
    if isinstance(node, TextNode):
        return

    for child in node.children:
        if not isinstance(child, ElementNode):
            continue

        if child.tag == "a":
            clean_link_node(request, result, child, force_shva)
            clean_links(request, result, child, force_shva)
        elif child.tag == "img":
            clean_image_node(request, result, child, force_shva)
        else:
            clean_links(request, result, child, force_shva)


def clean_link_node(
    request,
    result: dict,
    node: ElementNode,
    force_shva: bool,
):
    host = request.get_host()
    href = node.attrs.get("href") or "/"

    if is_internal_link(href, host):
        href = clean_internal_link(href, host)
        result["internal_links"].append(href)
        href = clean_attachment_link(href, force_shva)
    else:
        result["outgoing_links"].append(strip_link_protocol(href))
        href = assert_link_prefix(href)
        node.attrs["rel"] = "external nofollow noopener"

    node.attrs["target"] = "_blank"
    node.attrs["href"] = href

    if len(node.children) == 0:
        node.children.append(strip_link_protocol(href))
    elif len(node.children) == 1 and isinstance(node.children[0], TextNode):
        text = node.children[0].text
        if URL_RE.match(text):
            node.children[0].text = strip_link_protocol(text)


def clean_image_node(
    request,
    result: dict,
    node: ElementNode,
    force_shva: bool,
):
    host = request.get_host()
    src = node.attrs.get("src") or "/"

    node.attrs["alt"] = strip_link_protocol(node.attrs["alt"])

    if is_internal_link(src, host):
        src = clean_internal_link(src, host)
        result["images"].append(src)
        src = clean_attachment_link(src, force_shva)
    else:
        result["images"].append(strip_link_protocol(src))
        src = assert_link_prefix(src)

    node.attrs["src"] = src


def is_internal_link(link, host):
    if link.startswith("/") and not link.startswith("//"):
        return True

    link = strip_link_protocol(link).lstrip("www.").lower()
    return link.lower().startswith(host.lstrip("www."))


def strip_link_protocol(link):
    if link.lower().startswith("https:"):
        link = link[6:]
    if link.lower().startswith("http:"):
        link = link[5:]
    if link.startswith("//"):
        link = link[2:]
    return link


def assert_link_prefix(link):
    if link.lower().startswith("https:"):
        return link
    if link.lower().startswith("http:"):
        return link
    if link.startswith("//"):
        return "http:%s" % link

    return "http://%s" % link


def clean_internal_link(link, host):
    link = strip_link_protocol(link)

    if link.lower().startswith("www."):
        link = link[4:]
    if host.lower().startswith("www."):
        host = host[4:]

    if link.lower().startswith(host):
        link = link[len(host) :]

    return link or "/"


def clean_attachment_link(link, force_shva=False):
    try:
        resolution = resolve(link)
        if not resolution.namespaces:
            return link
        url_name = ":".join(resolution.namespaces + [resolution.url_name])
    except (Http404, ValueError):
        return link

    if url_name in MISAGO_ATTACHMENT_VIEWS:
        if force_shva:
            link = "%s?shva=1" % link
        elif link.endswith("?shva=1"):
            link = link[:-7]
    return link



from django import forms
from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied
from django.utils.translation import gettext_lazy as _

from ...acl import algebra
from ...acl.decorators import return_boolean
from ...acl.models import Role
from ...admin.forms import YesNoSwitch
from .decorators import authenticated_only

__all__ = [
    "allow_browse_users_list",
    "can_browse_users_list",
    "allow_follow_user",
    "can_follow_user",
    "allow_block_user",
    "can_block_user",
    "allow_see_ban_details",
    "can_see_ban_details",
]

CAN_BROWSE_USERS_LIST = YesNoSwitch(label=_("Can browse users list"), initial=1)
CAN_SEARCH_USERS = YesNoSwitch(label=_("Can search user profiles"), initial=1)
CAN_SEE_USER_NAME_HISTORY = YesNoSwitch(label=_("Can see other members name history"))
CAN_SEE_DETAILS = YesNoSwitch(
    label=_("Can see members bans details"),
    help_text=_(
        "Allows users with this permission to see user and staff ban messages."
    ),
)


class LimitedPermissionsForm(forms.Form):
    legend = _("User profiles")

    can_browse_users_list = CAN_BROWSE_USERS_LIST
    can_search_users = CAN_SEARCH_USERS
    can_see_users_name_history = CAN_SEE_USER_NAME_HISTORY
    can_see_ban_details = CAN_SEE_DETAILS


class PermissionsForm(LimitedPermissionsForm):
    can_browse_users_list = CAN_BROWSE_USERS_LIST
    can_search_users = CAN_SEARCH_USERS
    can_follow_users = YesNoSwitch(label=_("Can follow other users"), initial=1)
    can_be_blocked = YesNoSwitch(label=_("Can be blocked by other users"), initial=0)
    can_see_users_name_history = CAN_SEE_USER_NAME_HISTORY
    can_see_ban_details = CAN_SEE_DETAILS
    can_see_users_emails = YesNoSwitch(label=_("Can see members e-mails"))
    can_see_users_ips = YesNoSwitch(label=_("Can see members IPs"))
    can_see_hidden_users = YesNoSwitch(
        label=_("Can see members that hide their presence")
    )


def change_permissions_form(role):
    if isinstance(role, Role):
        if role.special_role == "anonymous":
            return LimitedPermissionsForm
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {
        "can_browse_users_list": 0,
        "can_search_users": 0,
        "can_follow_users": 0,
        "can_be_blocked": 1,
        "can_see_users_name_history": 0,
        "can_see_ban_details": 0,
        "can_see_users_emails": 0,
        "can_see_users_ips": 0,
        "can_see_hidden_users": 0,
    }
    new_acl.update(acl)

    return algebra.sum_acls(
        new_acl,
        roles=roles,
        key=key_name,
        can_browse_users_list=algebra.greater,
        can_search_users=algebra.greater,
        can_follow_users=algebra.greater,
        can_be_blocked=algebra.lower,
        can_see_users_name_history=algebra.greater,
        can_see_ban_details=algebra.greater,
        can_see_users_emails=algebra.greater,
        can_see_users_ips=algebra.greater,
        can_see_hidden_users=algebra.greater,
    )


def add_acl_to_user(user_acl, target):
    target.acl["can_have_attitude"] = False
    target.acl["can_follow"] = can_follow_user(user_acl, target)
    target.acl["can_block"] = can_block_user(user_acl, target)

    mod_permissions = ("can_have_attitude", "can_follow", "can_block")

    for permission in mod_permissions:
        if target.acl[permission]:
            target.acl["can_have_attitude"] = True
            break


def register_with(registry):
    registry.acl_annotator(get_user_model(), add_acl_to_user)


def allow_browse_users_list(user_acl):
    if not user_acl["can_browse_users_list"]:
        raise PermissionDenied(_("You can't browse users list."))


can_browse_users_list = return_boolean(allow_browse_users_list)


@authenticated_only
def allow_follow_user(user_acl, target):
    if not user_acl["can_follow_users"]:
        raise PermissionDenied(_("You can't follow other users."))
    if user_acl["user_id"] == target.id:
        raise PermissionDenied(_("You can't add yourself to followed."))


can_follow_user = return_boolean(allow_follow_user)


@authenticated_only
def allow_block_user(user_acl, target):
    if target.is_staff or target.is_superuser:
        raise PermissionDenied(_("You can't block administrators."))
    if user_acl["user_id"] == target.id:
        raise PermissionDenied(_("You can't block yourself."))
    # FIXME: check if user has "can be blocked" permission


can_block_user = return_boolean(allow_block_user)


@authenticated_only
def allow_see_ban_details(user_acl, target):
    if not user_acl["can_see_ban_details"]:
        raise PermissionDenied(_("You can't see users bans details."))


can_see_ban_details = return_boolean(allow_see_ban_details)



from rest_framework.response import Response

from ....core.shortcuts import get_int_or_404
from ...viewmodels import (
    ForumThreads,
    PrivateThreads,
    PrivateThreadsCategory,
    ThreadsCategory,
    ThreadsRootCategory,
)


class ThreadsList:
    threads = None

    def __call__(self, request, **kwargs):
        start = get_int_or_404(request.query_params.get("start", 0))
        list_type = request.query_params.get("list", "all")
        category = self.get_category(request, pk=request.query_params.get("category"))
        threads = self.get_threads(request, category, list_type, start)

        return Response(self.get_response_json(request, category, threads)["THREADS"])

    def get_category(self, request, pk=None):
        raise NotImplementedError(
            "Threads list has to implement get_category(request, pk=None)"
        )

    def get_threads(self, request, category, list_type, start):
        return self.threads(  # pylint: disable=not-callable
            request, category, list_type, start
        )

    def get_response_json(self, request, category, threads):
        return threads.get_frontend_context()


class ForumThreadsList(ThreadsList):
    threads = ForumThreads

    def get_category(self, request, pk=None):
        if pk:
            return ThreadsCategory(request, pk=pk)
        return ThreadsRootCategory(request)


class PrivateThreadsList(ThreadsList):
    threads = PrivateThreads

    def get_category(self, request, pk=None):
        return PrivateThreadsCategory(request)


threads_list_endpoint = ForumThreadsList()
private_threads_list_endpoint = PrivateThreadsList()



from datetime import timedelta

from django.urls import reverse
from django.utils import timezone

from .. import test
from ..models import Post, Thread
from ..test import patch_category_acl
from .test_threads_api import ThreadsApiTestCase


class PostDeleteApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.post = test.reply_thread(self.thread, poster=self.user)

        self.api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.post.pk},
        )

    def test_delete_anonymous(self):
        """api validates if deleting user is authenticated"""
        self.logout_user()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    @patch_category_acl({"can_hide_posts": 1, "can_hide_own_posts": 1})
    def test_no_permission(self):
        """api validates permission to delete post"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't delete posts in this category."}
        )

    @patch_category_acl(
        {"can_hide_posts": 1, "can_hide_own_posts": 2, "post_edit_time": 0}
    )
    def test_delete_other_user_post_no_permission(self):
        """api valdiates if user can delete other users posts"""
        self.post.poster = None
        self.post.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete other users posts in this category."},
        )

    @patch_category_acl(
        {"can_hide_posts": 1, "can_hide_own_posts": 2, "post_edit_time": 0}
    )
    def test_delete_protected_post_no_permission(self):
        """api validates if user can delete protected post"""
        self.post.is_protected = True
        self.post.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This post is protected. You can't delete it."}
        )

    @patch_category_acl(
        {"can_hide_posts": 1, "can_hide_own_posts": 2, "post_edit_time": 1}
    )
    def test_delete_protected_post_after_edit_time(self):
        """api validates if user can delete delete post after edit time"""
        self.post.posted_on = timezone.now() - timedelta(minutes=10)
        self.post.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete posts that are older than 1 minute."},
        )

    @patch_category_acl(
        {
            "can_hide_posts": 0,
            "can_hide_own_posts": 2,
            "post_edit_time": 0,
            "can_close_threads": False,
        }
    )
    def test_delete_post_closed_thread_no_permission(self):
        """api valdiates if user can delete posts in closed threads"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't delete posts in it."},
        )

    @patch_category_acl(
        {
            "can_hide_posts": 0,
            "can_hide_own_posts": 2,
            "post_edit_time": 0,
            "can_close_threads": False,
        }
    )
    def test_delete_post_closed_category_no_permission(self):
        """api valdiates if user can delete posts in closed categories"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't delete posts in it."},
        )

    @patch_category_acl({"can_hide_posts": 2, "can_hide_own_posts": 2})
    def test_delete_first_post(self):
        """api disallows first post deletion"""
        api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.thread.first_post_id},
        )

        response = self.client.delete(api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't delete thread's first post."}
        )

    @patch_category_acl({"can_hide_posts": 2, "can_hide_own_posts": 2})
    def test_delete_best_answer(self):
        """api disallows best answer deletion"""
        self.thread.set_best_answer(self.user, self.post)
        self.thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't delete this post because its marked as best answer."},
        )

    @patch_category_acl(
        {"can_hide_posts": 0, "can_hide_own_posts": 2, "post_edit_time": 0}
    )
    def test_delete_owned_post(self):
        """api deletes owned thread post"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

        self.thread = Thread.objects.get(pk=self.thread.pk)

        self.assertNotEqual(self.thread.last_post_id, self.post.pk)
        with self.assertRaises(Post.DoesNotExist):
            self.thread.post_set.get(pk=self.post.pk)

    @patch_category_acl({"can_hide_posts": 2, "can_hide_own_posts": 0})
    def test_delete_post(self):
        """api deletes thread post"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

        self.thread = Thread.objects.get(pk=self.thread.pk)

        self.assertNotEqual(self.thread.last_post_id, self.post.pk)
        with self.assertRaises(Post.DoesNotExist):
            self.thread.post_set.get(pk=self.post.pk)


class EventDeleteApiTests(ThreadsApiTestCase):
    def setUp(self):
        super().setUp()

        self.event = test.reply_thread(self.thread, poster=self.user, is_event=True)

        self.api_link = reverse(
            "misago:api:thread-post-detail",
            kwargs={"thread_pk": self.thread.pk, "pk": self.event.pk},
        )

    def test_delete_anonymous(self):
        """api validates if deleting user is authenticated"""
        self.logout_user()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    @patch_category_acl(
        {"can_hide_posts": 2, "can_hide_own_posts": 0, "can_hide_events": 0}
    )
    def test_no_permission(self):
        """api validates permission to delete event"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't delete events in this category."}
        )

    @patch_category_acl(
        {
            "can_hide_posts": 2,
            "can_hide_own_posts": 0,
            "can_hide_events": 2,
            "can_close_threads": False,
        }
    )
    def test_delete_event_closed_thread_no_permission(self):
        """api valdiates if user can delete events in closed threads"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't delete events in it."},
        )

    @patch_category_acl(
        {"can_hide_posts": 2, "can_hide_events": 2, "can_close_threads": False}
    )
    def test_delete_event_closed_category_no_permission(self):
        """api valdiates if user can delete events in closed categories"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't delete events in it."},
        )

    @patch_category_acl({"can_hide_posts": 0, "can_hide_events": 2})
    def test_delete_event(self):
        """api differs posts from events"""
        response = self.client.delete(self.api_link)
        self.assertEqual(response.status_code, 200)

        self.thread = Thread.objects.get(pk=self.thread.pk)

        self.assertNotEqual(self.thread.last_post_id, self.event.pk)
        with self.assertRaises(Post.DoesNotExist):
            self.thread.post_set.get(pk=self.event.pk)



from functools import wraps

from .dynamicsettings import DynamicSettings


class override_dynamic_settings:
    def __init__(self, **settings):
        self._overrides = settings

    def __enter__(self):
        DynamicSettings.override_settings(self._overrides)

    def __exit__(self, *_):
        DynamicSettings.remove_overrides()

    def __call__(self, f):
        @wraps(f)
        def test_function_wrapper(*args, **kwargs):
            with self:
                return f(*args, **kwargs)

        return test_function_wrapper



import re

from django import forms
from django.utils.translation import gettext_lazy as _

from .base import ChangeSettingsForm

GOOGLE_SITE_VERIFICATION = re.compile(
    "^google-site-verification: google([0-9a-z]+)\.html$"
)


class ChangeAnalyticsSettingsForm(ChangeSettingsForm):
    settings = ["google_tracking_id", "google_site_verification"]

    google_tracking_id = forms.CharField(
        label=_("Tracking ID"),
        help_text=_(
            "Setting the Tracking ID will result in gtag.js file being included in "
            "your site's HTML markup, enabling Google Analytics integration."
        ),
        required=False,
    )
    google_site_verification = forms.CharField(
        label=_("Site verification token"),
        help_text=_(
            "This token was extracted from uploaded site verification file. "
            "To change it, upload new verification file."
        ),
        required=False,
        disabled=True,
    )
    google_site_verification_file = forms.FileField(
        label=_("Upload site verification file"),
        help_text=_(
            "Site verification file can be downloaded from Search Console's "
            '"Ownership verification" page.'
        ),
        required=False,
    )

    def clean_google_site_verification_file(self):
        upload = self.cleaned_data.get("google_site_verification_file")
        if not upload:
            return None

        if upload.content_type != "text/html":
            raise forms.ValidationError(_("Submitted file type is not HTML."))

        file_content = upload.read().decode("utf-8")
        content_match = GOOGLE_SITE_VERIFICATION.match(file_content)
        if not content_match:
            raise forms.ValidationError(
                _("Submitted file doesn't contain a verification code.")
            )

        return content_match.group(1)

    def clean(self):
        cleaned_data = super().clean()
        if cleaned_data.get("google_site_verification_file"):
            new_verification = cleaned_data.pop("google_site_verification_file")
            cleaned_data["google_site_verification"] = new_verification
        return cleaned_data



import json

from django.urls import reverse

from .. import test
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...readtracker import poststracker
from ...users.test import AuthenticatedUserTestCase
from ..models import Post
from ..test import patch_category_acl, patch_other_category_acl


class ThreadPostSplitApiTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)
        self.posts = [
            test.reply_thread(self.thread).pk,
            test.reply_thread(self.thread).pk,
        ]

        self.api_link = reverse(
            "misago:api:thread-post-split", kwargs={"thread_pk": self.thread.pk}
        )

        Category(name="Other category", slug="other-category").insert_at(
            self.category, position="last-child", save=True
        )
        self.other_category = Category.objects.get(slug="other-category")

    def test_anonymous_user(self):
        """you need to authenticate to split posts"""
        self.logout_user()

        response = self.client.post(
            self.api_link, json.dumps({}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "This action is not available to guests."}
        )

    @patch_category_acl({"can_move_posts": False})
    def test_no_permission(self):
        """api validates permission to split"""
        response = self.client.post(
            self.api_link, json.dumps({}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(), {"detail": "You can't split posts from this thread."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_empty_data(self):
        """api handles empty data"""
        response = self.client.post(self.api_link)
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to split."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_data(self):
        """api handles post that is invalid type"""
        response = self.client.post(
            self.api_link, "[]", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {
                "non_field_errors": [
                    "Invalid data. Expected a dictionary, but got list."
                ]
            },
        )

        response = self.client.post(
            self.api_link, "123", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"non_field_errors": ["Invalid data. Expected a dictionary, but got int."]},
        )

        response = self.client.post(
            self.api_link, '"string"', content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"non_field_errors": ["Invalid data. Expected a dictionary, but got str."]},
        )

        response = self.client.post(
            self.api_link, "malformed", content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "JSON parse error - Expecting value: line 1 column 1 (char 0)"},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_no_posts_ids(self):
        """api rejects no posts ids"""
        response = self.client.post(
            self.api_link, json.dumps({}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to split."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_empty_posts_ids(self):
        """api rejects empty posts ids list"""
        response = self.client.post(
            self.api_link, json.dumps({"posts": []}), content_type="application/json"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You have to specify at least one post to split."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_posts_data(self):
        """api handles invalid data"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": "string"}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": 'Expected a list of items but got type "str".'}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_invalid_posts_ids(self):
        """api handles invalid post id"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [1, 2, "string"]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "One or more post ids received were invalid."}
        )

    @override_dynamic_settings(posts_per_page=5, posts_per_page_orphans=3)
    @patch_category_acl({"can_move_posts": True})
    def test_split_limit(self):
        """api rejects more posts than split limit"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": list(range(9))}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "No more than 8 posts can be split at a single time."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_invisible(self):
        """api validates posts visibility"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {"posts": [test.reply_thread(self.thread, is_unapproved=True).pk]}
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to split could not be found."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_event(self):
        """api rejects events split"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [test.reply_thread(self.thread, is_event=True).pk]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(response.json(), {"detail": "Events can't be split."})

    @patch_category_acl({"can_move_posts": True})
    def test_split_first_post(self):
        """api rejects first post split"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [self.thread.first_post_id]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(), {"detail": "You can't split thread's first post."}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_hidden_posts(self):
        """api recjects attempt to split urneadable hidden post"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [test.reply_thread(self.thread, is_hidden=True).pk]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "You can't split posts the content you can't see."},
        )

    @patch_category_acl({"can_move_posts": True, "can_close_threads": False})
    def test_split_posts_closed_thread_no_permission(self):
        """api recjects attempt to split posts from closed thread"""
        self.thread.is_closed = True
        self.thread.save()

        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [test.reply_thread(self.thread).pk]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "This thread is closed. You can't split posts in it."},
        )

    @patch_category_acl({"can_move_posts": True, "can_close_threads": False})
    def test_split_posts_closed_category_no_permission(self):
        """api recjects attempt to split posts from closed thread"""
        self.category.is_closed = True
        self.category.save()

        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [test.reply_thread(self.thread).pk]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "This category is closed. You can't split posts in it."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_other_thread_posts(self):
        """api recjects attempt to split other thread's post"""
        other_thread = test.post_thread(self.category)

        response = self.client.post(
            self.api_link,
            json.dumps({"posts": [test.reply_thread(other_thread, is_hidden=True).pk]}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.json(),
            {"detail": "One or more posts to split could not be found."},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_empty_new_thread_data(self):
        """api handles empty form data"""
        response = self.client.post(
            self.api_link,
            json.dumps({"posts": self.posts}),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "title": ["This field is required."],
                "category": ["This field is required."],
            },
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_invalid_final_title(self):
        """api rejects split because final thread title was invalid"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {"posts": self.posts, "title": "$$$", "category": self.category.id}
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_other_category_acl({"can_see": False})
    @patch_category_acl({"can_move_posts": True})
    def test_split_invalid_category(self):
        """api rejects split because final category was invalid"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.other_category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json, {"category": ["Requested category could not be found."]}
        )

    @patch_category_acl({"can_move_posts": True, "can_start_threads": False})
    def test_split_unallowed_start_thread(self):
        """api rejects split because category isn't allowing starting threads"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {"category": ["You can't create new threads in selected category."]},
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_invalid_weight(self):
        """api rejects split because final weight was invalid"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "weight": 4,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json, {"weight": ["Ensure this value is less than or equal to 2."]}
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split_unallowed_global_weight(self):
        """api rejects split because global weight was unallowed"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "weight": 2,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "weight": [
                    "You don't have permission to pin threads "
                    "globally in this category."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True, "can_pin_threads": 0})
    def test_split_unallowed_local_weight(self):
        """api rejects split because local weight was unallowed"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "weight": 1,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {"weight": ["You don't have permission to pin threads in this category."]},
        )

    @patch_category_acl({"can_move_posts": True, "can_pin_threads": 1})
    def test_split_allowed_local_weight(self):
        """api allows local weight"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 1,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True, "can_pin_threads": 2})
    def test_split_allowed_global_weight(self):
        """api allows global weight"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 2,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True, "can_close_threads": False})
    def test_split_unallowed_close(self):
        """api rejects split because closing thread was unallowed"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "is_closed": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "is_closed": [
                    "You don't have permission to close threads in this category."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True, "can_close_threads": True})
    def test_split_with_close(self):
        """api allows for closing thread"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 0,
                    "is_closed": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True, "can_hide_threads": 0})
    def test_split_unallowed_hidden(self):
        """api rejects split because hidden thread was unallowed"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Valid thread title",
                    "category": self.category.id,
                    "is_hidden": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "is_hidden": [
                    "You don't have permission to hide threads in this category."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True, "can_hide_threads": 1})
    def test_split_with_hide(self):
        """api allows for hiding thread"""
        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "$$$",
                    "category": self.category.id,
                    "weight": 0,
                    "is_hidden": True,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 400)

        response_json = response.json()
        self.assertEqual(
            response_json,
            {
                "title": [
                    "Thread title should be at least 5 characters long (it has 3)."
                ]
            },
        )

    @patch_category_acl({"can_move_posts": True})
    def test_split(self):
        """api splits posts to new thread"""
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 2)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Split thread.",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # thread was created
        split_thread = self.category.thread_set.get(slug="split-thread")
        self.assertEqual(split_thread.replies, 1)

        # posts were removed from old thread
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 0)

        # posts were moved to new thread
        self.assertEqual(split_thread.post_set.filter(pk__in=self.posts).count(), 2)

    @patch_category_acl({"can_move_posts": True})
    def test_split_best_answer(self):
        """api splits best answer to new thread"""
        best_answer = test.reply_thread(self.thread)

        self.thread.set_best_answer(self.user, best_answer)
        self.thread.synchronize()
        self.thread.save()

        self.thread.refresh_from_db()
        self.assertEqual(self.thread.best_answer, best_answer)
        self.assertEqual(self.thread.replies, 3)

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": [best_answer.pk],
                    "title": "Split thread.",
                    "category": self.category.id,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # best_answer was moved and unmarked
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 2)
        self.assertIsNone(self.thread.best_answer)

        split_thread = self.category.thread_set.get(slug="split-thread")
        self.assertEqual(split_thread.replies, 0)
        self.assertIsNone(split_thread.best_answer)

    @patch_other_category_acl(
        {
            "can_start_threads": True,
            "can_close_threads": True,
            "can_hide_threads": True,
            "can_pin_threads": 2,
        }
    )
    @patch_category_acl({"can_move_posts": True})
    def test_split_kitchensink(self):
        """api splits posts with kitchensink"""
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 2)

        poststracker.save_read(self.user, self.thread.first_post)
        for post in self.posts:
            poststracker.save_read(
                self.user, Post.objects.select_related().get(pk=post)
            )

        response = self.client.post(
            self.api_link,
            json.dumps(
                {
                    "posts": self.posts,
                    "title": "Split thread",
                    "category": self.other_category.id,
                    "weight": 2,
                    "is_closed": 1,
                    "is_hidden": 1,
                }
            ),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        # thread was created
        split_thread = self.other_category.thread_set.get(slug="split-thread")
        self.assertEqual(split_thread.replies, 1)
        self.assertEqual(split_thread.weight, 2)
        self.assertTrue(split_thread.is_closed)
        self.assertTrue(split_thread.is_hidden)

        # posts were removed from old thread
        self.thread.refresh_from_db()
        self.assertEqual(self.thread.replies, 0)

        # posts were moved to new thread
        self.assertEqual(split_thread.post_set.filter(pk__in=self.posts).count(), 2)

        # postreads were removed
        postreads = self.user.postread_set.filter(post__is_event=False).order_by("id")

        postreads_threads = list(postreads.values_list("thread_id", flat=True))
        self.assertEqual(postreads_threads, [self.thread.pk])

        postreads_categories = list(postreads.values_list("category_id", flat=True))
        self.assertEqual(postreads_categories, [self.category.pk])



from io import StringIO

from django.core.management import call_command

from ...threads.models import Thread
from ..management.commands import createfakethreads


def test_management_command_creates_fake_threads(db):
    call_command(createfakethreads.Command(), stdout=StringIO())
    assert Thread.objects.exists()



# Generated by Django 1.11.17 on 2019-01-08 19:49
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_core", "0002_basic_settings")]

    operations = [migrations.DeleteModel(name="CacheVersion")]



from django.conf import settings
from django.db import models
from django.utils import timezone


class PostLike(models.Model):
    category = models.ForeignKey("misago_categories.Category", on_delete=models.CASCADE)
    thread = models.ForeignKey("misago_threads.Thread", on_delete=models.CASCADE)
    post = models.ForeignKey("misago_threads.Post", on_delete=models.CASCADE)

    liker = models.ForeignKey(
        settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL
    )
    liker_name = models.CharField(max_length=255, db_index=True)
    liker_slug = models.CharField(max_length=255)

    liked_on = models.DateTimeField(default=timezone.now)

    class Meta:
        ordering = ["-id"]



from django import template
from django.utils.translation import gettext as _

register = template.Library()


@register.simple_tag
def pagetitle(title, **kwargs):
    if "page" in kwargs and kwargs["page"] > 1:
        title += " (%s)" % (_("page: %(page)s") % {"page": kwargs["page"]})

    if "parent" in kwargs:
        title += " | %s" % kwargs["parent"]

    return title



# Generated by Django 1.11.15 on 2018-08-16 17:29
import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0014_datadownload")]

    operations = [
        migrations.AddField(
            model_name="user",
            name="agreements",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.PositiveIntegerField(), default=list, size=None
            ),
        )
    ]



from django.utils import timezone

from .. import test
from ...categories.models import Category
from ...conf.test import override_dynamic_settings
from ...readtracker.poststracker import save_read
from ...users.test import AuthenticatedUserTestCase
from ..test import patch_category_acl

GOTO_URL = "%s#post-%s"
GOTO_PAGE_URL = "%s%s/#post-%s"

POSTS_PER_PAGE = 7
POSTS_PER_PAGE_ORPHANS = 3


class GotoViewTestCase(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)


class GotoPostTests(GotoViewTestCase):
    def test_goto_first_post(self):
        """first post redirect url is valid"""
        response = self.client.get(self.thread.first_post.get_absolute_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_URL % (self.thread.get_absolute_url(), self.thread.first_post_id),
        )

        response = self.client.get(response["location"])
        self.assertContains(response, self.thread.first_post.get_absolute_url())

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_last_post_on_page(self):
        """last post on page redirect url is valid"""
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS - 1):
            post = test.reply_thread(self.thread)

        response = self.client.get(post.get_absolute_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"], GOTO_URL % (self.thread.get_absolute_url(), post.pk)
        )

        response = self.client.get(response["location"])
        self.assertContains(response, post.get_absolute_url())

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_first_post_on_next_page(self):
        """first post on next page redirect url is valid"""
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS):
            post = test.reply_thread(self.thread)

        response = self.client.get(post.get_absolute_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 2, post.pk),
        )

        response = self.client.get(response["location"])
        self.assertContains(response, post.get_absolute_url())

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_first_post_on_page_three_out_of_five(self):
        """first post on next page redirect url is valid"""
        posts = []
        for _ in range(POSTS_PER_PAGE * 4 - 1):
            post = test.reply_thread(self.thread)
            posts.append(post)

        post = posts[POSTS_PER_PAGE * 2 - 3]

        response = self.client.get(post.get_absolute_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 3, post.pk),
        )

        response = self.client.get(response["location"])
        self.assertContains(response, post.get_absolute_url())

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_first_event_on_page_three_out_of_five(self):
        """event redirect url is valid"""
        posts = []
        for _ in range(POSTS_PER_PAGE * 4 - 1):
            post = test.reply_thread(self.thread)
            posts.append(post)

        post = posts[POSTS_PER_PAGE * 2 - 2]

        self.thread.has_events = True
        self.thread.save()

        post.is_event = True
        post.save()

        response = self.client.get(post.get_absolute_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 3, post.pk),
        )

        response = self.client.get(response["location"])
        self.assertContains(response, post.get_absolute_url())


class GotoLastTests(GotoViewTestCase):
    def test_goto_last_post(self):
        """first post redirect url is valid"""
        response = self.client.get(self.thread.get_last_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_URL % (self.thread.get_absolute_url(), self.thread.first_post_id),
        )

        response = self.client.get(response["location"])
        self.assertContains(response, self.thread.last_post.get_absolute_url())

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_last_post_on_page(self):
        """last post on page redirect url is valid"""
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS - 1):
            post = test.reply_thread(self.thread)

        response = self.client.get(self.thread.get_last_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"], GOTO_URL % (self.thread.get_absolute_url(), post.pk)
        )

        response = self.client.get(response["location"])
        self.assertContains(response, post.get_absolute_url())


class GotoNewTests(GotoViewTestCase):
    def test_goto_first_post(self):
        """first unread post redirect url is valid"""
        response = self.client.get(self.thread.get_new_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_URL % (self.thread.get_absolute_url(), self.thread.first_post_id),
        )

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_first_new_post(self):
        """first unread post redirect url in already read thread is valid"""
        save_read(self.user, self.thread.first_post)

        post = test.reply_thread(self.thread, posted_on=timezone.now())
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS - 1):
            test.reply_thread(self.thread, posted_on=timezone.now())

        response = self.client.get(self.thread.get_new_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"], GOTO_URL % (self.thread.get_absolute_url(), post.pk)
        )

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_first_new_post_on_next_page(self):
        """first unread post redirect url in already read multipage thread is valid"""
        save_read(self.user, self.thread.first_post)

        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS):
            last_post = test.reply_thread(self.thread, posted_on=timezone.now())
            save_read(self.user, last_post)

        post = test.reply_thread(self.thread, posted_on=timezone.now())
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS - 1):
            test.reply_thread(self.thread, posted_on=timezone.now())

        response = self.client.get(self.thread.get_new_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 2, post.pk),
        )

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_goto_first_new_post_in_read_thread(self):
        """goto new in read thread points to last post"""
        save_read(self.user, self.thread.first_post)

        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS):
            post = test.reply_thread(self.thread, posted_on=timezone.now())
            save_read(self.user, post)

        response = self.client.get(self.thread.get_new_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 2, post.pk),
        )

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_guest_goto_first_new_post_in_thread(self):
        """guest goto new in read thread points to last post"""
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS):
            post = test.reply_thread(self.thread, posted_on=timezone.now())

        self.logout_user()

        response = self.client.get(self.thread.get_new_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 2, post.pk),
        )


class GotoBestAnswerTests(GotoViewTestCase):
    def test_view_handles_no_best_answer(self):
        """if thread has no best answer, redirect to first post"""
        response = self.client.get(self.thread.get_best_answer_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_URL % (self.thread.get_absolute_url(), self.thread.first_post_id),
        )

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    def test_view_handles_best_answer(self):
        """if thread has best answer, redirect to it"""
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS):
            test.reply_thread(self.thread, posted_on=timezone.now())

        best_answer = test.reply_thread(self.thread, posted_on=timezone.now())
        self.thread.set_best_answer(self.user, best_answer)
        self.thread.save()

        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS - 1):
            test.reply_thread(self.thread, posted_on=timezone.now())

        response = self.client.get(self.thread.get_best_answer_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 2, best_answer.pk),
        )


class GotoUnapprovedTests(GotoViewTestCase):
    def test_view_validates_permission(self):
        """view validates permission to see unapproved posts"""
        response = self.client.get(self.thread.get_unapproved_post_url())
        self.assertContains(
            response, "You need permission to approve content", status_code=403
        )

        with patch_category_acl({"can_approve_content": True}):
            response = self.client.get(self.thread.get_unapproved_post_url())
            self.assertEqual(response.status_code, 302)

    @patch_category_acl({"can_approve_content": True})
    def test_view_handles_no_unapproved_posts(self):
        """if thread has no unapproved posts, redirect to last post"""
        response = self.client.get(self.thread.get_unapproved_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_URL % (self.thread.get_absolute_url(), self.thread.first_post_id),
        )

    @override_dynamic_settings(
        posts_per_page=POSTS_PER_PAGE, posts_per_page_orphans=POSTS_PER_PAGE_ORPHANS
    )
    @patch_category_acl({"can_approve_content": True})
    def test_view_handles_unapproved_posts(self):
        """if thread has unapproved posts, redirect to first of them"""
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS):
            test.reply_thread(self.thread, posted_on=timezone.now())

        post = test.reply_thread(
            self.thread, is_unapproved=True, posted_on=timezone.now()
        )
        for _ in range(POSTS_PER_PAGE + POSTS_PER_PAGE_ORPHANS - 1):
            test.reply_thread(self.thread, posted_on=timezone.now())

        response = self.client.get(self.thread.get_unapproved_post_url())
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            response["location"],
            GOTO_PAGE_URL % (self.thread.get_absolute_url(), 2, post.pk),
        )


class ThreadGotoPostTests(GotoViewTestCase):
    """brureforcing regression tests for regression test for #869"""

    def test_thread_growing_post_goto(self):
        """growing thread goto views don't fail"""
        for _ in range(60):
            post = test.reply_thread(self.thread, posted_on=timezone.now())

            # go to post link is valid
            post_url = self.client.get(post.get_absolute_url())["location"]

            response = self.client.get(post_url)
            self.assertContains(response, post.get_absolute_url())

            # go to last post link is valid
            last_url = self.client.get(self.thread.get_last_post_url())["location"]
            self.assertEqual(post_url, last_url)

    def test_thread_growing_event_goto(self):
        """growing thread goto views don't fail for events"""
        for i in range(60):
            test.reply_thread(self.thread, posted_on=timezone.now())

            post = test.reply_thread(self.thread, posted_on=timezone.now())
            post.is_event = True
            post.save()

            # go to post link is valid
            post_url = self.client.get(post.get_absolute_url())["location"]

            if i == 0:
                # manually set events flag after first event was created
                self.thread.has_events = True
                self.thread.save()

            response = self.client.get(post_url)
            self.assertContains(response, post.get_absolute_url())

            # go to last post link is valid
            last_url = self.client.get(self.thread.get_last_post_url())["location"]
            self.assertEqual(post_url, last_url)

    def test_thread_post_goto(self):
        """thread goto views don't fail"""
        for _ in range(60):
            test.reply_thread(self.thread, posted_on=timezone.now())

        for post in self.thread.post_set.order_by("id").iterator():
            # go to post link is valid
            post_url = self.client.get(post.get_absolute_url())["location"]

            response = self.client.get(post_url)
            self.assertContains(response, post.get_absolute_url())

        # go to last post link is valid
        last_url = self.client.get(self.thread.get_last_post_url())["location"]
        self.assertEqual(post_url, last_url)

    def test_thread_event_goto(self):
        """thread goto views don't fail for events"""
        for _ in range(60):
            test.reply_thread(self.thread, posted_on=timezone.now())

            post = test.reply_thread(self.thread, posted_on=timezone.now())
            post.is_event = True
            post.save()

        for post in (
            self.thread.post_set.filter(is_event=True).order_by("id").iterator()
        ):
            # go to post link is valid
            post_url = self.client.get(post.get_absolute_url())["location"]

            response = self.client.get(post_url)
            self.assertContains(response, post.get_absolute_url())

        # go to last post link is valid
        last_url = self.client.get(self.thread.get_last_post_url())["location"]
        self.assertEqual(post_url, last_url)



"""
Misago-native rehash of Django's createsuperuser command that
works with double authentication fields on user model
"""
import sys
from getpass import getpass

from django.contrib.auth import get_user_model
from django.contrib.auth.password_validation import validate_password
from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand
from django.db import DEFAULT_DB_ALIAS, IntegrityError
from django.utils.encoding import force_str

from ....conf.shortcuts import get_dynamic_settings
from ...setupnewuser import setup_new_user
from ...validators import validate_email, validate_username

User = get_user_model()


class NotRunningInTTYException(Exception):
    pass


class Command(BaseCommand):
    help = "Used to create a superuser."

    def add_arguments(self, parser):
        parser.add_argument(
            "--username",
            dest="username",
            default=None,
            help="Specifies the username for the superuser.",
        )
        parser.add_argument(
            "--email",
            dest="email",
            default=None,
            help="Specifies the e-mail for the superuser.",
        )
        parser.add_argument(
            "--password",
            dest="password",
            default=None,
            help="Specifies the password for the superuser.",
        )
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            default=True,
            help=(
                "Tells Misago to NOT prompt the user for input "
                "of any kind. You must use --username with "
                "--noinput, along with an option for any other "
                "required field. Superusers created with "
                "--noinput will  not be able to log in until "
                "they're given a valid password."
            ),
        )
        parser.add_argument(
            "--database",
            action="store",
            dest="database",
            default=DEFAULT_DB_ALIAS,
            help=('Specifies the database to use. Default is "default".'),
        )

    def execute(self, *args, **options):
        self.stdin = options.get("stdin", sys.stdin)  # Used for testing
        return super().execute(*args, **options)

    def handle(
        self, *args, **options
    ):  # pylint: disable=too-many-branches, too-many-locals
        username = options.get("username")
        email = options.get("email")
        password = options.get("password")
        interactive = options.get("interactive")
        verbosity = int(options.get("verbosity", 1))

        settings = get_dynamic_settings()

        # Validate initial inputs
        if username is not None:
            try:
                username = username.strip()
                validate_username(settings, username)
            except ValidationError as e:
                self.stderr.write("\n".join(e.messages))
                username = None

        if email is not None:
            try:
                email = email.strip()
                validate_email(email)
            except ValidationError as e:
                self.stderr.write("\n".join(e.messages))
                email = None

        if password is not None:
            password = password.strip()
            if password == "":
                self.stderr.write("Error: Blank passwords aren't allowed.")

        if not interactive:
            if username and email and password:
                # Call User manager's create_superuser using our wrapper
                self.create_superuser(username, email, password, settings, verbosity)
        else:
            try:
                if hasattr(self.stdin, "isatty") and not self.stdin.isatty():
                    raise NotRunningInTTYException("Not running in a TTY")

                # Prompt for username/password, and any other required fields.
                # Enclose this whole thing in a try/except to trap for a
                # keyboard interrupt and exit gracefully.
                while not username:
                    try:
                        message = force_str("Enter displayed username: ")
                        raw_value = input(message).strip()
                        validate_username(settings, raw_value)
                        username = raw_value
                    except ValidationError as e:
                        self.stderr.write("\n".join(e.messages))

                while not email:
                    try:
                        raw_value = input("Enter e-mail address: ").strip()
                        validate_email(raw_value)
                        email = raw_value
                    except ValidationError as e:
                        self.stderr.write("\n".join(e.messages))

                while not password:
                    raw_value = getpass("Enter password: ")
                    password_repeat = getpass("Repeat password:")
                    if raw_value != password_repeat:
                        self.stderr.write("Error: Your passwords didn't match.")
                        # Don't validate passwords that don't match.
                        continue
                    if raw_value.strip() == "":
                        self.stderr.write("Error: Blank passwords aren't allowed.")
                        # Don't validate blank passwords.
                        continue
                    try:
                        validate_password(
                            raw_value, user=User(username=username, email=email)
                        )
                    except ValidationError as e:
                        self.stderr.write("\n".join(e.messages))
                        response = input(
                            "Bypass password validation and create user anyway? [y/N]: "
                        )
                        if response.lower() != "y":
                            continue
                    password = raw_value

                # Call User manager's create_superuser using our wrapper
                self.create_superuser(username, email, password, settings, verbosity)

            except KeyboardInterrupt:
                self.stderr.write("\nOperation cancelled.")
                sys.exit(1)
            except NotRunningInTTYException:
                self.stdout.write(
                    "Superuser creation skipped due to not running in a TTY. "
                    "You can run `manage.py createsuperuser` in your project "
                    "to create one manually."
                )

    def create_superuser(self, username, email, password, settings, verbosity):
        try:
            user = User.objects.create_superuser(username, email, password)
            setup_new_user(settings, user)

            if verbosity >= 1:
                message = "Superuser #%s has been created successfully."
                self.stdout.write(message % user.pk)
        except ValidationError as e:
            self.stderr.write(e.messages[0])
        except IntegrityError as e:
            self.stderr.write(e.messages[0])



from django import forms
from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.utils.translation import gettext_lazy as _

from ..acl import algebra
from ..acl.decorators import return_boolean
from ..admin.forms import YesNoSwitch
from .models import Category, CategoryRole, RoleCategoryACL


class PermissionsForm(forms.Form):
    legend = _("Category access")

    can_see = YesNoSwitch(label=_("Can see category"))
    can_browse = YesNoSwitch(label=_("Can see category contents"))


def change_permissions_form(role):
    if isinstance(role, CategoryRole):
        return PermissionsForm


def build_acl(acl, roles, key_name):
    new_acl = {"visible_categories": [], "browseable_categories": [], "categories": {}}
    new_acl.update(acl)

    roles = get_categories_roles(roles)

    for category in Category.objects.all_categories():
        build_category_acl(new_acl, category, roles, key_name)

    return new_acl


def get_categories_roles(roles):
    queryset = RoleCategoryACL.objects.filter(role__in=roles)
    queryset = queryset.select_related("category_role")

    roles = {}
    for acl_relation in queryset.iterator():
        role = acl_relation.category_role
        roles.setdefault(acl_relation.category_id, []).append(role)
    return roles


def build_category_acl(acl, category, categories_roles, key_name):
    if category.level > 1:
        if category.parent_id not in acl["visible_categories"]:
            # dont bother with child categories of invisible parents
            return
        if not acl["categories"][category.parent_id]["can_browse"]:
            # parent's visible, but its contents aint
            return

    category_roles = categories_roles.get(category.pk, [])

    final_acl = {"can_see": 0, "can_browse": 0}

    algebra.sum_acls(
        final_acl,
        roles=category_roles,
        key=key_name,
        can_see=algebra.greater,
        can_browse=algebra.greater,
    )

    if final_acl["can_see"]:
        acl["visible_categories"].append(category.pk)
        acl["categories"][category.pk] = final_acl

        if final_acl["can_browse"]:
            acl["browseable_categories"].append(category.pk)


def add_acl_to_category(user_acl, target):
    target.acl["can_see"] = can_see_category(user_acl, target)
    target.acl["can_browse"] = can_browse_category(user_acl, target)


def serialize_categories_acls(user_acl):
    categories_acl = []
    for category, acl in user_acl.pop("categories").items():
        if acl["can_browse"]:
            categories_acl.append(
                {
                    "id": category,
                    "can_start_threads": acl.get("can_start_threads", False),
                    "can_reply_threads": acl.get("can_reply_threads", False),
                    "can_pin_threads": acl.get("can_pin_threads", 0),
                    "can_hide_threads": acl.get("can_hide_threads", 0),
                    "can_close_threads": acl.get("can_close_threads", False),
                }
            )
    user_acl["categories"] = categories_acl


def register_with(registry):
    registry.acl_annotator(Category, add_acl_to_category)
    registry.user_acl_serializer(serialize_categories_acls)


def allow_see_category(user_acl, target):
    try:
        category_id = target.pk
    except AttributeError:
        category_id = int(target)

    if not category_id in user_acl["visible_categories"]:
        raise Http404()


can_see_category = return_boolean(allow_see_category)


def allow_browse_category(user_acl, target):
    target_acl = user_acl["categories"].get(target.id, {"can_browse": False})
    if not target_acl["can_browse"]:
        message = _('You don\'t have permission to browse "%(category)s" contents.')
        raise PermissionDenied(message % {"category": target.name})


can_browse_category = return_boolean(allow_browse_category)



from .models import Subscription


def make_subscription_aware(user, target):
    if hasattr(target, "__iter__"):
        make_threads_subscription_aware(user, target)
    else:
        make_thread_subscription_aware(user, target)


def make_threads_subscription_aware(user, threads):
    if not threads:
        return

    if user.is_anonymous:
        for thread in threads:
            thread.subscription = None
    else:
        threads_dict = {}
        for thread in threads:
            thread.subscription = None
            threads_dict[thread.pk] = thread

        subscriptions_queryset = user.subscription_set.filter(
            thread_id__in=threads_dict.keys()
        )

        for subscription in subscriptions_queryset.iterator():
            threads_dict[subscription.thread_id].subscription = subscription


def make_thread_subscription_aware(user, thread):
    if user.is_anonymous:
        thread.subscription = None
    else:
        try:
            thread.subscription = user.subscription_set.get(thread=thread)
        except Subscription.DoesNotExist:
            thread.subscription = None



from ..activepostersranking import get_active_posters_ranking
from ..online.utils import make_users_status_aware
from ..serializers import UserCardSerializer


class ActivePosters:
    def __init__(self, request):
        ranking = get_active_posters_ranking()
        make_users_status_aware(request, ranking["users"], fetch_state=True)

        self.count = ranking["users_count"]
        self.tracked_period = request.settings.top_posters_ranking_length
        self.users = ranking["users"]

    def get_frontend_context(self):
        return {
            "tracked_period": self.tracked_period,
            "results": ScoredUserSerializer(self.users, many=True).data,
            "count": self.count,
        }

    def get_template_context(self):
        return {
            "tracked_period": self.tracked_period,
            "users": self.users,
            "users_count": self.count,
        }


ScoredUserSerializer = UserCardSerializer.extend_fields("meta")



from io import StringIO

from django.core.management import call_command

from ..management.commands import createfakefollowers
from ..users import get_fake_user


def test_management_command_creates_fake_followers_for_two_users(user, other_user):
    call_command(createfakefollowers.Command(), stdout=StringIO())


def test_management_command_creates_fake_followers_for_multiple_users(db, fake):
    [get_fake_user(fake) for i in range(10)]
    call_command(createfakefollowers.Command(), stdout=StringIO())


def test_management_command_displays_error_if_no_users_exist(db):
    stderr = StringIO()
    call_command(createfakefollowers.Command(), stderr=stderr)
    stderr.seek(0)
    assert stderr.read()


def test_management_command_displays_error_if_only_one_user_exist(user):
    stderr = StringIO()
    call_command(createfakefollowers.Command(), stderr=stderr)
    stderr.seek(0)
    assert stderr.read()



from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ....core.pgutils import chunk_queryset

User = get_user_model()


class Command(BaseCommand):
    help = "Deletes specified profile field from database."

    def add_arguments(self, parser):
        parser.add_argument("fieldname", help="field to delete", nargs="?")

    def handle(self, *args, **options):
        fieldname = options["fieldname"]
        if not fieldname:
            self.stderr.write("Specify fieldname to delete.")
            return

        fields_deleted = 0

        queryset = User.objects.filter(profile_fields__has_keys=[fieldname])

        for user in chunk_queryset(queryset):
            if fieldname in user.profile_fields.keys():
                user.profile_fields.pop(fieldname)
                user.save(update_fields=["profile_fields"])
                fields_deleted += 1

        self.stdout.write(
            '"%s" profile field has been deleted from %s users.'
            % (fieldname, fields_deleted)
        )



from django.core.exceptions import PermissionDenied
from django.utils.module_loading import import_string

from ..conf import settings


class SearchProviders:
    def __init__(self, search_providers):
        self._initialized = False
        self._providers = []

        self.providers = search_providers

    def initialize_providers(self):
        if self._initialized:
            return
        self._initialized = True

        self._providers = list(map(import_string, self.providers))

    def get_providers(self, request):
        if not self._initialized:
            self.initialize_providers()

        providers = []
        for provider in self._providers:
            providers.append(provider(request))
        return providers

    def get_allowed_providers(self, request):
        allowed_providers = []
        for provider in self.get_providers(request):
            try:
                provider.allow_search()
                allowed_providers.append(provider)
            except PermissionDenied:
                pass
        return allowed_providers


searchproviders = SearchProviders(settings.MISAGO_SEARCH_EXTENSIONS)



from .analytics import ChangeAnalyticsSettingsForm
from .base import ChangeSettingsForm
from .captcha import ChangeCaptchaSettingsForm
from .general import ChangeGeneralSettingsForm
from .oauth2 import ChangeOAuth2SettingsForm
from .threads import ChangeThreadsSettingsForm
from .users import ChangeUsersSettingsForm



from ...test import assert_contains, assert_not_contains
from ...threads.test import post_thread
from ..models import Setting
from ..test import override_dynamic_settings


@override_dynamic_settings(forum_address="http://test.com/")
def test_default_og_image_is_used_when_none_is_set(db, client):
    response = client.get("/")
    assert_contains(response, "http://test.com/static/misago/img/og-image.jpg")


@override_dynamic_settings(forum_address="http://test.com/")
def test_custom_og_image_is_used_instead_of_default_one_when_set(db, client):
    Setting.objects.filter(setting="og_image").update(
        image="custom-image.jpg", image_width=600, image_height=300
    )

    response = client.get("/")
    assert_not_contains(response, "http://test.com/media/misago/img/og-image.jpg")
    assert_contains(response, "http://test.com/media/custom-image.jpg")
    assert_contains(response, 'property="og:image:width" content="600"')
    assert_contains(response, 'property="og:image:height" content="300"')


@override_dynamic_settings(forum_address="http://test.com/")
def test_default_og_image_is_used_on_user_profiles(client, user):
    response = client.get("%sposts/" % user.get_absolute_url())
    assert_contains(response, "http://test.com/static/misago/img/og-image.jpg")


@override_dynamic_settings(
    forum_address="http://test.com/", og_image_avatar_on_profile=True
)
def test_user_avatar_can_be_used_as_og_image_on_user_profiles(client, user):
    response = client.get("%sposts/" % user.get_absolute_url())
    assert_not_contains(response, "http://test.com/static/misago/img/og-image.jpg")


@override_dynamic_settings(forum_address="http://test.com/")
def test_default_og_image_is_used_on_thread_page(client, default_category, user):
    thread = post_thread(default_category, poster=user)
    response = client.get(thread.get_absolute_url())
    assert_contains(response, "http://test.com/static/misago/img/og-image.jpg")


@override_dynamic_settings(
    forum_address="http://test.com/", og_image_avatar_on_thread=True
)
def test_thread_started_avatar_can_be_used_as_og_image_on_thread_page(
    client, default_category, user
):
    thread = post_thread(default_category, poster=user)
    response = client.get(thread.get_absolute_url())
    assert_not_contains(response, "http://test.com/static/misago/img/og-image.jpg")



from django import forms
from django.utils.translation import gettext_lazy as _

from ....admin.forms import YesNoSwitch
from .base import ChangeSettingsForm


class ChangeCaptchaSettingsForm(ChangeSettingsForm):
    settings = [
        "captcha_type",
        "recaptcha_site_key",
        "recaptcha_secret_key",
        "qa_question",
        "qa_help_text",
        "qa_answers",
        "enable_stop_forum_spam",
        "stop_forum_spam_confidence",
    ]

    captcha_type = forms.ChoiceField(
        label=_("Enable CAPTCHA"),
        choices=[
            ("no", _("No CAPTCHA")),
            ("re", _("reCaptcha")),
            ("qa", _("Question and answer")),
        ],
        widget=forms.RadioSelect(),
    )

    recaptcha_site_key = forms.CharField(
        label=_("Site key"), max_length=100, required=False
    )
    recaptcha_secret_key = forms.CharField(
        label=_("Secret key"), max_length=100, required=False
    )

    qa_question = forms.CharField(
        label=_("Test question"), max_length=100, required=False
    )
    qa_help_text = forms.CharField(
        label=_("Question help text"), max_length=250, required=False
    )
    qa_answers = forms.CharField(
        label=_("Valid answers"),
        help_text=_("Enter each answer in new line. Answers are case-insensitive."),
        widget=forms.Textarea({"rows": 4}),
        max_length=250,
        required=False,
    )

    enable_stop_forum_spam = YesNoSwitch(
        label=_("Validate new registrations against SFS database"),
        help_text=_(
            "Turning this option on will result in Misago validating new user's e-mail "
            "and IP address against SFS database."
        ),
    )
    stop_forum_spam_confidence = forms.IntegerField(
        label=_("Minimum SFS confidence required"),
        help_text=_(
            "SFS compares user e-mail and IP address with database of known spammers "
            "and assigns the confidence score in range of 0 to 100 that user is a "
            "spammer themselves. If this score is equal or higher than specified, "
            "Misago will block user from registering and ban their IP address "
            "for 24 hours."
        ),
        min_value=0,
        max_value=100,
    )

    def clean(self):
        cleaned_data = super().clean()

        if cleaned_data.get("captcha_type") == "re":
            if not cleaned_data.get("recaptcha_site_key"):
                self.add_error(
                    "recaptcha_site_key",
                    _(
                        "You need to enter site key if "
                        "selected CAPTCHA type is reCaptcha."
                    ),
                )

            if not cleaned_data.get("recaptcha_secret_key"):
                self.add_error(
                    "recaptcha_secret_key",
                    _(
                        "You need to enter secret key if "
                        "selected CAPTCHA type is reCaptcha."
                    ),
                )

        if cleaned_data.get("captcha_type") == "qa":
            if not cleaned_data.get("qa_question"):
                self.add_error(
                    "qa_question",
                    _("You need to set question if selected CAPTCHA type is Q&A."),
                )

            if not cleaned_data.get("qa_answers"):
                self.add_error(
                    "qa_answers",
                    _(
                        "You need to set question answers if "
                        "selected CAPTCHA type is Q&A."
                    ),
                )

        return cleaned_data



# Generated by Django 3.2.15 on 2023-01-04 12:37

from django.db import migrations

settings = (
    "enable_sso",
    "sso_public_key",
    "sso_private_key",
    "sso_url",
)


def delete_sso_settings(apps, _):
    Setting = apps.get_model("misago_conf", "Setting")
    Setting.objects.filter(setting__in=settings).delete()


class Migration(migrations.Migration):

    dependencies = [
        ("misago_conf", "0007_add_oauth2_settings"),
    ]

    operations = [migrations.RunPython(delete_sso_settings)]



from django.contrib import messages
from django.shortcuts import redirect
from django.utils.translation import gettext as _

from ...admin.views import render
from ..models import Icon
from .forms import IconsForm


def icons_admin(request):
    form = IconsForm()
    if request.method == "POST":
        form = IconsForm(request.POST, request.FILES)
        if form.is_valid():
            form.save()

            messages.success(request, _("Icons have been updated."))
            return redirect("misago:admin:settings:icons:index")

    return render(
        request, "misago/admin/icons.html", {"form": form, "icons": get_icons()}
    )


def get_icons():
    return {i.type: i for i in Icon.objects.all()}



import pytest
from django.urls import reverse

from ...models import MenuItem


@pytest.fixture
def list_url(admin_client):
    return reverse("misago:admin:settings:menu-items:index")


@pytest.fixture
def menu_item(superuser):
    return MenuItem.objects.create(
        menu=MenuItem.MENU_NAVBAR,
        title="Test TMLA",
        url="https://top_menu_item_admin.com",
        order=0,
    )


@pytest.fixture
def other_menu_item(superuser):
    return MenuItem.objects.create(
        menu=MenuItem.MENU_BOTH,
        title="Other Menu Item",
        url="https://other_menu_item.com",
        order=1,
    )



from unittest.mock import Mock

from ..applyfilters import apply_filters


def test_filtering_calls_action_with_given_args():
    action = Mock()
    apply_filters(action, [], 1, 2)
    action.assert_called_once_with(1, 2)


def test_filtering_calls_action_with_given_kwargs():
    action = Mock()
    apply_filters(action, [], a=1, b=2)
    action.assert_called_once_with(a=1, b=2)


def test_filtering_returns_action_return_value():
    action = Mock(return_value=42)
    assert apply_filters(action, []) == 42


def test_filter_is_called_with_action_as_first_arg():
    filter_ = Mock()
    action = Mock()
    apply_filters(action, [filter_])
    filter_.assert_called_once_with(action)


def test_filter_is_called_with_given_args():
    filter_ = Mock()
    action = Mock()
    apply_filters(action, [filter_], 1, 2)
    filter_.assert_called_once_with(action, 1, 2)


def test_filter_is_called_with_given_kwargs():
    filter_ = Mock()
    action = Mock()
    apply_filters(action, [filter_], a=1, b=2)
    filter_.assert_called_once_with(action, a=1, b=2)


def test_filters_are_chained():
    action = Mock(return_value=1)

    def multiply_filter(action):
        return action() * 2

    assert apply_filters(action, [multiply_filter, multiply_filter]) == 4



from unittest.mock import Mock

import pytest
from django import forms
from django.template import Context, Template, TemplateSyntaxError

from ..forms import YesNoSwitch
from ..templatetags.misago_admin_form import (
    get_field_image_dimensions,
    is_multiple_choice_field,
    is_radio_select_field,
    is_select_field,
    is_textarea_field,
    is_yesno_switch_field,
    render_attrs,
    render_bool_attrs,
)


class Form(forms.Form):
    text_field = forms.CharField(
        label="Hello!", max_length=255, help_text="I am a help text."
    )
    textarea_field = forms.CharField(
        label="Message", max_length=255, widget=forms.Textarea()
    )
    select_field = forms.ChoiceField(
        label="Choice", choices=(("y", "Yes"), ("n", "No"))
    )
    checkbox_select_field = forms.MultipleChoiceField(
        label="Color",
        choices=(("r", "Red"), ("g", "Green"), ("b", "Blue")),
        widget=forms.CheckboxSelectMultiple,
    )
    multiple_select_field = forms.MultipleChoiceField(
        label="Rank", choices=(("r", "Red"), ("g", "Green"), ("b", "Blue"))
    )
    yesno_field = YesNoSwitch(label="Switch")
    image_field = forms.ImageField(label="Image!", help_text="I am a help text.")


@pytest.fixture
def form():
    return Form()


def render(template_str, form=None):
    base_template = "{%% load misago_admin_form %%} %s"
    context = Context({"form": form or Form()})
    template = Template(base_template % template_str)
    return template.render(context).strip()


def test_row_with_field_input_is_rendered():
    html = render("{% form_row form.text_field %}")
    assert "id_text_field" in html


def test_row_with_field_input_and_label_css_class_is_rendered():
    html = render('{% form_row form.text_field label_class="col-md-3" %}')
    assert "id_text_field" in html
    assert "col-md-3" in html


def test_row_with_field_input_and_field_css_class_is_rendered():
    html = render('{% form_row form.text_field field_class="col-md-9" %}')
    assert "id_text_field" in html
    assert "col-md-9" in html


def test_row_with_field_input_and_label_and_field_css_classes_is_rendered():
    html = render('{% form_row form.text_field "col-md-3" "col-md-9" %}')
    assert "id_text_field" in html
    assert "col-md-3" in html
    assert "col-md-9" in html


def test_field_label_is_rendered():
    html = render("{% form_row form.text_field %}")
    assert "Hello!" in html


def test_field_help_text_is_rendered():
    html = render("{% form_row form.text_field %}")
    assert "I am a help text." in html


def test_tag_without_field_raises_exception():
    with pytest.raises(TemplateSyntaxError):
        render("{% form_row %}")


def test_image_row_with_field_input_is_rendered():
    html = render("{% form_image_row form.image_field %}")
    assert "id_image_field" in html


def test_image_row_with_field_input_and_label_css_class_is_rendered():
    html = render('{% form_image_row form.image_field label_class="col-md-3" %}')
    assert "id_image_field" in html
    assert "col-md-3" in html


def test_image_row_with_field_input_and_field_css_class_is_rendered():
    html = render('{% form_image_row form.image_field field_class="col-md-9" %}')
    assert "id_image_field" in html
    assert "col-md-9" in html


def test_image_row_with_field_input_and_label_and_field_css_classes_is_rendered():
    html = render('{% form_image_row form.image_field "col-md-3" "col-md-9" %}')
    assert "id_image_field" in html
    assert "col-md-3" in html
    assert "col-md-9" in html


def test_image_field_label_is_rendered():
    html = render("{% form_image_row form.image_field %}")
    assert "Image!" in html


def test_image_field_help_text_is_rendered():
    html = render("{% form_image_row form.image_field %}")
    assert "I am a help text." in html


class TestImage:
    url = "test-image.png"


def render_image(template_str):
    base_template = "{%% load misago_admin_form %%} %s"
    context = Context({"form": Form(initial={"image_field": TestImage()})})
    template = Template(base_template % template_str)
    return template.render(context).strip()


def test_image_row_with_value_renders_label():
    html = render_image("{% form_image_row form.image_field %}")
    assert "Image!" in html


def test_image_row_with_value_renders_help_text():
    html = render_image("{% form_image_row form.image_field %}")
    assert "I am a help text." in html


def test_image_row_with_value_renders_image_preview():
    html = render_image("{% form_image_row form.image_field %}")
    assert ('src="%s"' % TestImage.url) in html


def test_image_row_with_value_renders_input():
    html = render_image("{% form_image_row form.image_field %}")
    assert "id_image_field" in html


def test_image_row_with_value_renders_input_and_label_css_class():
    html = render_image('{% form_image_row form.image_field label_class="col-md-3" %}')
    assert "id_image_field" in html
    assert "col-md-3" in html


def test_image_row_with_value_renders_input_and_field_css_class():
    html = render_image('{% form_image_row form.image_field field_class="col-md-9" %}')
    assert "id_image_field" in html
    assert "col-md-9" in html


def test_image_row_with_value_renders_input_and_label_and_field_css_classes():
    html = render_image('{% form_image_row form.image_field "col-md-3" "col-md-9" %}')
    assert "id_image_field" in html
    assert "col-md-3" in html
    assert "col-md-9" in html


def test_image_tag_without_field_raises_exception():
    with pytest.raises(TemplateSyntaxError):
        render("{% form_image_row %}")


def test_for_field_with_radio_select_widget_filter_returns_true(form):
    assert is_radio_select_field(form["yesno_field"])


def test_for_field_without_radio_select_widget_filter_returns_false(form):
    assert not is_radio_select_field(form["text_field"])


def test_for_field_with_select_widget_filter_returns_true(form):
    assert is_select_field(form["select_field"])


def teste_for_field_without_select_widget_filter_returns_false(form):
    assert not is_select_field(form["text_field"])


def test_for_field_with_checkbox_select_widget_filter_returns_true(form):
    assert is_multiple_choice_field(form["checkbox_select_field"])


def test_for_field_without_checkbox_select_widget_filter_returns_false(form):
    assert not is_multiple_choice_field(form["text_field"])


def test_for_field_with_multiple_select_widget_filter_returns_true(form):
    assert is_multiple_choice_field(form["multiple_select_field"])


def test_for_field_without_multiple_select_widget_filter_returns_false(form):
    assert not is_multiple_choice_field(form["text_field"])


def test_for_field_with_textarea_widget_filter_returns_true(form):
    assert is_textarea_field(form["textarea_field"])


def test_for_field_without_textarea_widget_filter_returns_false(form):
    assert not is_textarea_field(form["text_field"])


def test_for_yes_no_field_filter_returns_true(form):
    assert is_yesno_switch_field(form["yesno_field"])


def test_for_non_yes_no_field_filter_returns_false(form):
    assert not is_yesno_switch_field(form["text_field"])


def test_specified_class_name_is_rendered():
    result = render_attrs({"attrs": {}}, class_name="form-control")
    assert result == 'class="form-control"'


def test_specified_class_name_overrided_by_class_attr():
    result = render_attrs({"attrs": {"class": "custom"}}, class_name="form-control")
    assert result == 'class="custom"'


def test_attr_with_string_value_is_rendered():
    result = render_attrs({"attrs": {"name": "lorem"}})
    assert result == 'name="lorem"'


def test_attr_with_int_value_is_rendered():
    result = render_attrs({"attrs": {"cols": 5}})
    assert result == 'cols="5"'


def test_attr_with_boolean_true_value_is_not_rendered():
    result = render_attrs({"attrs": {"selected": True}})
    assert result == ""


def test_attr_with_boolean_false_value_is_not_rendered():
    result = render_attrs({"attrs": {"selected": False}})
    assert result == ""


def test_attr_with_none_value_is_not_rendered():
    result = render_attrs({"attrs": {"selected": None}})
    assert result == ""


def test_attr_name_is_escaped():
    result = render_attrs({"attrs": {'"': "test"}})
    assert result == '&quot;="test"'


def test_attr_value_is_escaped():
    result = render_attrs({"attrs": {"name": '"'}})
    assert result == 'name="&quot;"'


def test_multiple_valid_attrs_are_rendered():
    result = render_attrs({"attrs": {"name": "lorem", "cols": 5}})
    assert result == 'name="lorem" cols="5"'


def test_empty_attr_dict_is_not_rendered():
    result = render_attrs({"attrs": {}})
    assert result == ""


def test_attr_with_boolean_true_value_is_rendered():
    result = render_bool_attrs({"bool": True})
    assert result == "bool"


def test_bool_attr_with_string_value_is_not_rendered():
    result = render_bool_attrs({"name": "hello"})
    assert result == ""


def test_bool_attr_with_int_value_is_not_rendered():
    result = render_bool_attrs({"col": 13})
    assert result == ""


def test_bool_attr_with_boolean_false_value_is_not_rendered():
    result = render_bool_attrs({"selected": False})
    assert result == ""


def test_bool_attr_with_none_value_is_not_rendered():
    result = render_bool_attrs({"selected": None})
    assert result == ""


def test_bool_attr_with_false_int_value_is_not_rendered():
    result = render_bool_attrs({"selected": 0})
    assert result == ""


def test_multiple_bool_attrs_with_boolean_true_value_are_rendered():
    result = render_bool_attrs({"selected": True, "required": True})
    assert result == "selected required"


def test_only_bool_attrs_with_boolean_true_value_are_rendered():
    result = render_bool_attrs({"bool": True, "string": "hello", "int": 123})
    assert result == "bool"


def test_util_turns_dimensions_tuple_into_dict():
    assert get_field_image_dimensions((20, 40)) == {"width": 20, "height": 40}


def test_util_returs_none_if_dimensions_are_empty():
    assert get_field_image_dimensions(None) is None



from django.urls import reverse

from .. import test
from ...acl.test import patch_user_acl
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase


class FloodProtectionTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(category=self.category)

        self.post_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

    def test_flood_has_no_showstoppers(self):
        """endpoint handles posting interruption"""
        response = self.client.post(
            self.post_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            self.post_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.json(),
            {"detail": "You can't post message so quickly after previous one."},
        )

    @patch_user_acl({"can_omit_flood_protection": True})
    def test_user_with_permission_omits_flood_protection(self):
        response = self.client.post(
            self.post_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        response = self.client.post(
            self.post_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)



from datetime import timedelta

from django.core import mail
from django.urls import reverse
from django.utils import timezone
from django.utils.encoding import smart_str

from ...conf.test import override_dynamic_settings
from .. import test
from ...categories.models import Category
from ...users.test import AuthenticatedUserTestCase, create_test_user
from ..test import patch_category_acl, patch_other_user_category_acl


class EmailNotificationTests(AuthenticatedUserTestCase):
    def setUp(self):
        super().setUp()

        self.category = Category.objects.get(slug="first-category")
        self.thread = test.post_thread(
            category=self.category, started_on=timezone.now() - timedelta(seconds=5)
        )

        self.api_link = reverse(
            "misago:api:thread-post-list", kwargs={"thread_pk": self.thread.pk}
        )

        self.other_user = create_test_user("OtherUser", "otheruser@example.com")

    @patch_category_acl({"can_reply_threads": True})
    def test_no_subscriptions(self):
        """no emails are sent because noone subscibes to thread"""
        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    @patch_category_acl({"can_reply_threads": True})
    def test_poster_not_notified(self):
        """no emails are sent because only poster subscribes to thread"""
        self.user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=timezone.now(),
            send_email=True,
        )

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    @patch_category_acl({"can_reply_threads": True})
    def test_other_user_no_email_subscription(self):
        """no emails are sent because subscriber has e-mails off"""
        self.other_user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=timezone.now(),
            send_email=False,
        )

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    @patch_category_acl({"can_reply_threads": True})
    @patch_other_user_category_acl({"can_see": False})
    def test_other_user_no_permission(self):
        """no emails are sent because subscriber has no permission to read thread"""
        self.other_user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=timezone.now(),
            send_email=True,
        )

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    @patch_category_acl({"can_reply_threads": True})
    def test_moderation_queue(self):
        """no emails are sent because new post is moderated"""
        self.category.require_replies_approval = True
        self.category.save()

        self.other_user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=timezone.now(),
            send_email=True,
        )

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    @patch_category_acl({"can_reply_threads": True})
    def test_other_user_not_read(self):
        """no emails are sent because subscriber didn't read previous post"""
        self.other_user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=timezone.now(),
            send_email=True,
        )

        test.reply_thread(self.thread, posted_on=timezone.now())

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 0)

    @override_dynamic_settings(forum_address="http://test.com/")
    @patch_category_acl({"can_reply_threads": True})
    def test_other_notified(self):
        """email is sent to subscriber"""
        self.other_user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=timezone.now(),
            send_email=True,
        )

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 1)
        last_email = mail.outbox[-1]

        self.assertIn(self.user.username, last_email.subject)
        self.assertIn(self.thread.title, last_email.subject)

        message = smart_str(last_email.body)

        self.assertIn(self.user.username, message)
        self.assertIn(self.thread.title, message)
        self.assertIn(self.thread.get_absolute_url(), message)

        last_post = self.thread.post_set.order_by("id").last()
        self.assertIn(last_post.get_absolute_url(), message)

    @override_dynamic_settings(forum_address="http://test.com/")
    @patch_category_acl({"can_reply_threads": True})
    def test_other_notified_after_reading(self):
        """email is sent to subscriber that had sub updated by read api"""
        self.other_user.subscription_set.create(
            thread=self.thread,
            category=self.category,
            last_read_on=self.thread.last_post_on,
            send_email=True,
        )

        response = self.client.post(
            self.api_link, data={"post": "This is test response!"}
        )
        self.assertEqual(response.status_code, 200)

        self.assertEqual(len(mail.outbox), 1)
        last_email = mail.outbox[-1]

        self.assertIn(self.user.username, last_email.subject)
        self.assertIn(self.thread.title, last_email.subject)

        message = smart_str(last_email.body)

        self.assertIn(self.user.username, message)
        self.assertIn(self.thread.title, message)
        self.assertIn(self.thread.get_absolute_url(), message)

        last_post = self.thread.post_set.order_by("id").last()
        self.assertIn(last_post.get_absolute_url(), message)



# Generated by Django 1.11.1 on 2017-05-26 21:56
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("misago_users", "0008_ban_registration_only")]

    operations = []



from django.contrib import messages
from django.contrib.auth import get_user_model, update_session_auth_hash
from django.db import transaction
from django.http import JsonResponse
from django.shortcuts import redirect
from django.utils.translation import gettext_lazy as _

from ....acl.useracl import get_user_acl
from ....admin.auth import authorize_admin
from ....admin.views import generic
from ....categories.models import Category
from ....core.mail import mail_users
from ....core.pgutils import chunk_queryset
from ....threads.models import Thread
from ...avatars.dynamic import set_avatar as set_dynamic_avatar
from ...datadownloads import request_user_data_download, user_has_data_download_request
from ...deletesrecord import record_user_deleted_by_staff
from ...models import Ban
from ...profilefields import profilefields
from ...setupnewuser import setup_new_user
from ...signatures import set_user_signature
from ..forms import (
    BanUsersForm,
    EditUserForm,
    EditUserFormFactory,
    NewUserForm,
    create_filter_users_form,
)
from ..tasks import delete_user_with_content

User = get_user_model()


class UserAdmin(generic.AdminBaseMixin):
    root_link = "misago:admin:users:index"
    templates_dir = "misago/admin/users"
    model = User

    def get_form_class(self, request, target):
        add_is_active_fields = False
        add_admin_fields = False

        if not target.is_deleting_account:
            if not target.is_staff:
                add_is_active_fields = True
            elif request.user.is_superuser:
                add_is_active_fields = request.user.pk != target.pk

        if request.user.is_superuser:
            add_admin_fields = request.user.pk != target.pk

        return EditUserFormFactory(
            self.form_class,
            target,
            add_is_active_fields=add_is_active_fields,
            add_admin_fields=add_admin_fields,
        )


class UsersList(UserAdmin, generic.ListView):
    items_per_page = 24
    ordering = [
        ("-id", _("From newest")),
        ("id", _("From oldest")),
        ("slug", _("A to z")),
        ("-slug", _("Z to a")),
        ("-posts", _("Biggest posters")),
        ("posts", _("Smallest posters")),
    ]
    selection_label = _("With users: 0")
    empty_selection_label = _("Select users")
    mass_actions = [
        {"action": "activate", "name": _("Activate accounts")},
        {"action": "ban", "name": _("Ban users"), "icon": "fa fa-lock"},
        {"action": "request_data_download", "name": _("Request data download")},
        {
            "action": "delete_accounts",
            "name": _("Delete accounts"),
            "confirmation": _("Are you sure you want to delete selected users?"),
        },
        {
            "action": "delete_all",
            "name": _("Delete with content"),
            "confirmation": _(
                "Are you sure you want to delete selected users? "
                "This will also delete all content associated with their accounts."
            ),
            "is_atomic": False,
        },
    ]

    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related("rank")

    def get_filter_form(self, request):
        return create_filter_users_form()

    def action_activate(self, request, users):
        inactive_users = []
        for user in users:
            if user.requires_activation:
                inactive_users.append(user)

        if not inactive_users:
            message = _("You have to select inactive users.")
            raise generic.MassActionError(message)
        else:
            activated_users_pks = [u.pk for u in inactive_users]
            queryset = User.objects.filter(pk__in=activated_users_pks)
            queryset.update(requires_activation=User.ACTIVATION_NONE)

            subject = _("Your account on %(forum_name)s forums has been activated")
            mail_subject = subject % {"forum_name": request.settings.forum_name}

            mail_users(
                inactive_users,
                mail_subject,
                "misago/emails/activation/by_admin",
                context={"settings": request.settings},
            )

            messages.success(request, _("Selected users accounts have been activated."))

    def action_ban(
        self, request, users
    ):  # pylint: disable=too-many-locals, too-many-nested-blocks, too-many-branches
        users = users.order_by("slug")
        for user in users:
            if user.is_superuser:
                message = _("%(user)s is super admin and can't be banned.")
                mesage = message % {"user": user.username}
                raise generic.MassActionError(mesage)

        form = BanUsersForm(users=users)
        if "finalize" in request.POST:
            form = BanUsersForm(request.POST, users=users)
            if form.is_valid():
                cleaned_data = form.cleaned_data
                banned_values = []

                ban_kwargs = {
                    "user_message": cleaned_data.get("user_message"),
                    "staff_message": cleaned_data.get("staff_message"),
                    "expires_on": cleaned_data.get("expires_on"),
                }

                for user in users:
                    for ban in cleaned_data["ban_type"]:
                        banned_value = None

                        if ban == "usernames":
                            check_type = Ban.USERNAME
                            banned_value = user.username.lower()

                        if ban == "emails":
                            check_type = Ban.EMAIL
                            banned_value = user.email.lower()

                        if ban == "domains":
                            check_type = Ban.EMAIL
                            banned_value = user.email.lower()
                            at_pos = banned_value.find("@")
                            banned_value = "*%s" % banned_value[at_pos:]

                        if ban == "ip" and user.joined_from_ip:
                            check_type = Ban.IP
                            banned_value = user.joined_from_ip

                        if ban in ("ip_first", "ip_two") and user.joined_from_ip:
                            check_type = Ban.IP

                            if ":" in user.joined_from_ip:
                                ip_separator = ":"
                            if "." in user.joined_from_ip:
                                ip_separator = "."

                            bits = user.joined_from_ip.split(ip_separator)
                            if ban == "ip_first":
                                formats = (bits[0], ip_separator)
                            if ban == "ip_two":
                                formats = (bits[0], ip_separator, bits[1], ip_separator)
                            banned_value = "%s*" % ("".join(formats))

                        if banned_value and banned_value not in banned_values:
                            ban_kwargs.update(
                                {"check_type": check_type, "banned_value": banned_value}
                            )
                            Ban.objects.create(**ban_kwargs)
                            banned_values.append(banned_value)

                Ban.objects.invalidate_cache()
                messages.success(request, _("Selected users have been banned."))
                return None

        return self.render(
            request,
            {"users": users, "form": form},
            template_name="misago/admin/users/ban.html",
        )

    def action_request_data_download(self, request, users):
        for user in users:
            if not user_has_data_download_request(user):
                request_user_data_download(user, requester=request.user)

        messages.success(
            request, _("Data download requests have been placed for selected users.")
        )

    def action_delete_accounts(self, request, users):
        for user in users:
            if user == request.user:
                raise generic.MassActionError(_("You can't delete yourself."))
            if user.is_staff or user.is_superuser:
                message = _("%(user)s is admin and can't be deleted.") % {
                    "user": user.username
                }
                raise generic.MassActionError(message)

        for user in users:
            user.delete(anonymous_username=request.settings.anonymous_username)
            record_user_deleted_by_staff()

        messages.success(request, _("Selected users have been deleted."))

    def action_delete_all(self, request, users):
        for user in users:
            if user == request.user:
                raise generic.MassActionError(_("You can't delete yourself."))
            if user.is_staff or user.is_superuser:
                message = _("%(user)s is admin and can't be deleted.") % {
                    "user": user.username
                }
                raise generic.MassActionError(message)

        for user in users:
            user.is_active = False
            user.save()

            delete_user_with_content.delay(user.pk)

        messages.success(
            request,
            _(
                "Selected users have been disabled and queued for deletion "
                "together with their content."
            ),
        )


class NewUser(UserAdmin, generic.ModelFormView):
    form_class = NewUserForm
    template_name = "new.html"
    message_submit = _('New user "%(user)s" has been registered.')

    def get_form(self, form_class, request, target):
        if request.method == "POST":
            return form_class(
                request.POST, request.FILES, instance=target, request=request
            )
        return form_class(instance=target, request=request)

    def handle_form(self, form, request, target):
        new_user = User.objects.create_user(
            form.cleaned_data["username"],
            form.cleaned_data["email"],
            form.cleaned_data["new_password"],
            title=form.cleaned_data["title"],
            rank=form.cleaned_data.get("rank"),
            joined_from_ip=request.user_ip,
        )

        if form.cleaned_data.get("roles"):
            new_user.roles.add(*form.cleaned_data["roles"])

        new_user.update_acl_key()
        setup_new_user(request.settings, new_user)

        messages.success(request, self.message_submit % {"user": target.username})
        return redirect("misago:admin:users:edit", pk=new_user.pk)


class EditUser(UserAdmin, generic.ModelFormView):
    form_class = EditUserForm
    template_name = "edit.html"
    message_submit = _('User "%(user)s" has been edited.')

    def real_dispatch(self, request, target):
        target.old_username = target.username
        target.old_is_avatar_locked = target.is_avatar_locked
        return super().real_dispatch(request, target)

    def get_form(self, form_class, request, target):
        if request.method == "POST":
            return form_class(
                request.POST, request.FILES, instance=target, request=request
            )
        return form_class(instance=target, request=request)

    def handle_form(self, form, request, target):
        target.username = target.old_username
        if target.username != form.cleaned_data.get("username"):
            target.set_username(
                form.cleaned_data.get("username"), changed_by=request.user
            )

        if form.cleaned_data.get("new_password"):
            target.set_password(form.cleaned_data["new_password"])

        if form.cleaned_data.get("email"):
            target.set_email(form.cleaned_data["email"])

        if form.cleaned_data.get("is_avatar_locked"):
            if not target.old_is_avatar_locked:
                set_dynamic_avatar(target)

        if "is_staff" in form.fields and "is_superuser" in form.fields:
            target.is_staff = form.cleaned_data.get("is_staff")
            target.is_superuser = form.cleaned_data.get("is_superuser")

        if "is_active" in form.fields and "is_active_staff_message" in form.fields:
            target.is_active = form.cleaned_data.get("is_active")
            target.is_active_staff_message = form.cleaned_data.get(
                "is_active_staff_message"
            )

        target.rank = form.cleaned_data.get("rank")

        target.roles.clear()
        target.roles.add(*form.cleaned_data["roles"])

        target_acl = get_user_acl(target, request.cache_versions)
        set_user_signature(
            request, target, target_acl, form.cleaned_data.get("signature")
        )

        profilefields.update_user_profile_fields(request, target, form)

        target.update_acl_key()
        target.save()

        if target.pk == request.user.pk:
            authorize_admin(request)
            update_session_auth_hash(request, target)

        messages.success(request, self.message_submit % {"user": target.username})



from unittest.mock import Mock

import pytest

from .. import THEME_CACHE
from ..context_processors import theme as context_processor


@pytest.fixture
def mock_request(cache_versions):
    return Mock(cache_versions=cache_versions)


def test_theme_data_is_included_in_template_context(db, mock_request):
    assert context_processor(mock_request)["theme"]


def test_theme_is_loaded_from_database_if_cache_is_not_available(
    db, mocker, mock_request, django_assert_num_queries
):
    mocker.patch("django.core.cache.cache.get", return_value=None)
    with django_assert_num_queries(3):
        context_processor(mock_request)


def test_theme_is_loaded_from_cache_if_it_is_set(
    db, mocker, mock_request, django_assert_num_queries
):
    cache_get = mocker.patch("django.core.cache.cache.get", return_value={})
    with django_assert_num_queries(0):
        context_processor(mock_request)
    cache_get.assert_called_once()


def test_theme_cache_is_set_if_none_exists(db, mocker, mock_request):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)

    context_processor(mock_request)
    cache_set.assert_called_once()


def test_theme_cache_is_not_set_if_it_already_exists(
    db, mocker, mock_request, django_assert_num_queries
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value={})
    with django_assert_num_queries(0):
        context_processor(mock_request)
    cache_set.assert_not_called()


def test_theme_cache_key_includes_cache_name_and_version(
    db, mocker, mock_request, cache_versions
):
    cache_set = mocker.patch("django.core.cache.cache.set")
    mocker.patch("django.core.cache.cache.get", return_value=None)
    context_processor(mock_request)
    cache_key = cache_set.call_args[0][0]
    assert THEME_CACHE in cache_key
    assert cache_versions[THEME_CACHE] in cache_key



from datetime import timedelta

from django.test import TestCase
from django.utils import timezone

from .. import test
from ...categories.models import Category
from ...users.test import create_test_user
from ..models import Poll, Post, Thread, ThreadParticipant


class ThreadModelTests(TestCase):
    def setUp(self):
        datetime = timezone.now()

        self.category = Category.objects.all_categories()[:1][0]
        self.thread = Thread(
            category=self.category,
            started_on=datetime,
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=datetime,
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        self.thread.set_title("Test thread")
        self.thread.save()

        Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster_name="Tester",
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        self.thread.synchronize()
        self.thread.save()

    def test_synchronize(self):
        """synchronize method updates thread data to reflect its contents"""
        user = create_test_user("User", "user@example.com")

        self.assertEqual(self.thread.replies, 0)

        datetime = timezone.now() + timedelta(5)
        post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        # first sync call, updates last thread
        self.thread.synchronize()

        self.assertEqual(self.thread.last_post, post)
        self.assertEqual(self.thread.last_post_on, post.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)
        self.assertFalse(self.thread.has_reported_posts)
        self.assertFalse(self.thread.has_unapproved_posts)
        self.assertFalse(self.thread.has_hidden_posts)
        self.assertEqual(self.thread.replies, 1)

        # add unapproved post
        unapproved_post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime + timedelta(5),
            updated_on=datetime + timedelta(5),
            is_unapproved=True,
        )

        self.thread.synchronize()
        self.assertEqual(self.thread.last_post, post)
        self.assertEqual(self.thread.last_post_on, post.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)
        self.assertFalse(self.thread.has_reported_posts)
        self.assertTrue(self.thread.has_unapproved_posts)
        self.assertFalse(self.thread.has_hidden_posts)
        self.assertEqual(self.thread.replies, 1)

        # add hidden post
        hidden_post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime + timedelta(10),
            updated_on=datetime + timedelta(10),
            is_hidden=True,
        )

        self.thread.synchronize()
        self.assertEqual(self.thread.last_post, hidden_post)
        self.assertEqual(self.thread.last_post_on, hidden_post.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)
        self.assertFalse(self.thread.has_reported_posts)
        self.assertTrue(self.thread.has_unapproved_posts)
        self.assertTrue(self.thread.has_hidden_posts)
        self.assertEqual(self.thread.replies, 2)

        # unhide post
        hidden_post.is_hidden = False
        hidden_post.save()

        # last post changed to unhidden one
        self.thread.synchronize()
        self.assertEqual(self.thread.last_post, hidden_post)
        self.assertEqual(self.thread.last_post_on, hidden_post.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)
        self.assertFalse(self.thread.has_reported_posts)
        self.assertTrue(self.thread.has_unapproved_posts)
        self.assertFalse(self.thread.has_hidden_posts)
        self.assertEqual(self.thread.replies, 2)

        # unmoderate post
        unapproved_post.is_unapproved = False
        unapproved_post.save()

        # last post not changed, but flags and count did
        self.thread.synchronize()
        self.assertEqual(self.thread.last_post, hidden_post)
        self.assertEqual(self.thread.last_post_on, hidden_post.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)
        self.assertFalse(self.thread.has_reported_posts)
        self.assertFalse(self.thread.has_unapproved_posts)
        self.assertFalse(self.thread.has_hidden_posts)
        self.assertEqual(self.thread.replies, 3)

        # add event post
        event = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="-",
            parsed="-",
            checksum="nope",
            posted_on=datetime + timedelta(10),
            updated_on=datetime + timedelta(10),
            is_event=True,
        )

        self.thread.synchronize()
        self.assertEqual(self.thread.last_post, event)
        self.assertEqual(self.thread.last_post_on, event.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)
        self.assertTrue(self.thread.last_post_is_event)
        self.assertTrue(self.thread.has_events)
        self.assertFalse(self.thread.has_reported_posts)
        self.assertFalse(self.thread.has_unapproved_posts)
        self.assertFalse(self.thread.has_hidden_posts)
        # events don't count to reply count
        self.assertEqual(self.thread.replies, 3)

        # create another post to provoke other has_events resolution path
        Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        self.thread.synchronize()
        self.assertFalse(self.thread.last_post_is_event)
        self.assertTrue(self.thread.has_events)

        # remove event
        event.delete()

        self.thread.synchronize()
        self.assertFalse(self.thread.last_post_is_event)
        self.assertFalse(self.thread.has_events)

        # has poll flag
        self.assertFalse(self.thread.has_poll)

        Poll.objects.create(
            thread=self.thread,
            category=self.category,
            poster_name="test",
            poster_slug="test",
            choices=[],
        )

        self.thread.synchronize()
        self.assertTrue(self.thread.has_poll)

    def test_set_first_post(self):
        """set_first_post sets first post and poster data on thread"""
        user = create_test_user("User", "user@example.com")

        datetime = timezone.now() + timedelta(5)

        post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        self.thread.set_first_post(post)
        self.assertEqual(self.thread.first_post, post)
        self.assertEqual(self.thread.started_on, post.posted_on)
        self.assertEqual(self.thread.starter, user)
        self.assertEqual(self.thread.starter_name, user.username)
        self.assertEqual(self.thread.starter_slug, user.slug)

    def test_set_last_post(self):
        """set_last_post sets first post and poster data on thread"""
        user = create_test_user("User", "user@example.com")

        datetime = timezone.now() + timedelta(5)

        post = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        self.thread.set_last_post(post)
        self.assertEqual(self.thread.last_post, post)
        self.assertEqual(self.thread.last_post_on, post.posted_on)
        self.assertEqual(self.thread.last_poster, user)
        self.assertEqual(self.thread.last_poster_name, user.username)
        self.assertEqual(self.thread.last_poster_slug, user.slug)

    def test_set_best_answer(self):
        """set_best_answer sets best answer and setter data on thread"""
        user = create_test_user("User", "user@example.com")

        best_answer = Post.objects.create(
            category=self.category,
            thread=self.thread,
            poster=user,
            poster_name=user.username,
            original="Hello! I am test message!",
            parsed="<p>Hello! I am test message!</p>",
            checksum="nope",
            posted_on=timezone.now(),
            updated_on=timezone.now(),
            is_protected=True,
        )

        self.thread.synchronize()
        self.thread.save()

        self.thread.set_best_answer(user, best_answer)
        self.thread.save()

        self.assertEqual(self.thread.best_answer, best_answer)
        self.assertTrue(self.thread.has_best_answer)
        self.assertTrue(self.thread.best_answer_is_protected)
        self.assertTrue(self.thread.best_answer_marked_on)
        self.assertEqual(self.thread.best_answer_marked_by, user)
        self.assertEqual(self.thread.best_answer_marked_by_name, user.username)
        self.assertEqual(self.thread.best_answer_marked_by_slug, user.slug)

        # clear best answer
        self.thread.clear_best_answer()

        self.assertIsNone(self.thread.best_answer)
        self.assertFalse(self.thread.has_best_answer)
        self.assertFalse(self.thread.best_answer_is_protected)
        self.assertIsNone(self.thread.best_answer_marked_on)
        self.assertIsNone(self.thread.best_answer_marked_by)
        self.assertIsNone(self.thread.best_answer_marked_by_name)
        self.assertIsNone(self.thread.best_answer_marked_by_slug)

    def test_set_invalid_best_answer(self):
        """set_best_answer implements some assertions for data integrity"""
        user = create_test_user("User", "user@example.com")

        other_thread = test.post_thread(self.category)
        with self.assertRaises(ValueError):
            self.thread.set_best_answer(user, other_thread.first_post)

        with self.assertRaises(ValueError):
            self.thread.set_best_answer(user, self.thread.first_post)

        with self.assertRaises(ValueError):
            reply = test.reply_thread(self.thread, is_hidden=True)
            self.thread.set_best_answer(user, reply)

        with self.assertRaises(ValueError):
            reply = test.reply_thread(self.thread, is_unapproved=True)
            self.thread.set_best_answer(user, reply)

    def test_move(self):
        """move(new_category) moves thread to other category"""
        root_category = Category.objects.root_category()
        Category(name="New Category", slug="new-category").insert_at(
            root_category, position="last-child", save=True
        )
        new_category = Category.objects.get(slug="new-category")

        self.thread.move(new_category)
        self.assertEqual(self.thread.category, new_category)

        for post in self.thread.post_set.all():
            self.assertEqual(post.category_id, new_category.id)

    def test_merge(self):
        """merge(other_thread) moves other thread content to this thread"""
        with self.assertRaises(ValueError):
            self.thread.merge(self.thread)

        datetime = timezone.now() + timedelta(5)

        other_thread = Thread(
            category=self.category,
            started_on=datetime,
            starter_name="Tester",
            starter_slug="tester",
            last_post_on=datetime,
            last_poster_name="Tester",
            last_poster_slug="tester",
        )

        other_thread.set_title("Other thread")
        other_thread.save()

        post = Post.objects.create(
            category=self.category,
            thread=other_thread,
            poster_name="Admin",
            original="Hello! I am other message!",
            parsed="<p>Hello! I am other message!</p>",
            checksum="nope",
            posted_on=datetime,
            updated_on=datetime,
        )

        other_thread.first_post = post
        other_thread.last_post = post
        other_thread.save()

        self.thread.merge(other_thread)

        self.thread.synchronize()
        self.assertEqual(self.thread.replies, 1)
        self.assertEqual(self.thread.last_post, post)
        self.assertEqual(self.thread.last_post_on, post.posted_on)
        self.assertEqual(self.thread.last_poster_name, "Admin")
        self.assertEqual(self.thread.last_poster_slug, "admin")

    def test_delete_private_thread(self):
        """
        private thread gets deleted automatically
        when there are no participants left in it
        """
        user = create_test_user("User", "user@example.com")
        other_user = create_test_user("OtherUser", "otheruser@example.com")

        ThreadParticipant.objects.add_participants(self.thread, [user, other_user])
        self.assertEqual(self.thread.participants.count(), 2)

        user.delete(anonymous_username="Deleted")
        Thread.objects.get(id=self.thread.id)

        other_user.delete(anonymous_username="Deleted")
        with self.assertRaises(Thread.DoesNotExist):
            Thread.objects.get(id=self.thread.id)



import pytest
from django import forms
from django.urls import reverse

from ....cache.test import assert_invalidates_cache
from ... import SETTINGS_CACHE
from ..forms import ChangeSettingsForm
from ..views import ChangeSettingsView


@pytest.fixture(autouse=True)
def messages_mock(mocker):
    return mocker.patch("misago.conf.admin.views.messages")


class Form(ChangeSettingsForm):
    settings = ["forum_name"]

    forum_name = forms.CharField(max_length=255)


class View(ChangeSettingsView):
    form_class = Form

    def render(self, request, context):
        return True


def test_view_loads_form_settings_from_db(setting):
    view = View()
    assert view.get_settings(["forum_name"]) == {setting.setting: setting}


def test_view_raises_value_error_if_requested_setting_is_not_found(setting):
    view = View()
    with pytest.raises(ValueError):
        assert view.get_settings(["forum_name", "nonexisting_setting"])


def test_initial_form_data_is_build_from_settings_dict(setting):
    view = View()
    settings_dict = {setting.setting: setting}
    assert view.get_initial_form_data(settings_dict) == {setting.setting: setting.value}


def test_view_handles_get_request(rf, setting):
    view = View.as_view()
    view(rf.get("/"))


def test_view_handles_post_request(rf, setting):
    view = View.as_view()
    view(rf.post("/", {setting.setting: "New Value"}))


def test_view_changes_setting_on_correct_post_request(rf, setting):
    view = View.as_view()
    view(rf.post("/", {setting.setting: "New Value"}))

    setting.refresh_from_db()
    assert setting.value == "New Value"


def test_view_handles_invalid_post_request(rf, setting):
    view = View.as_view()
    view(rf.post("/", {"invalid_setting": ""}))

    setting.refresh_from_db()
    assert setting.value == "Misago"


def test_view_invalidates_settings_cache_on_correct_post_request(rf, setting):
    with assert_invalidates_cache(SETTINGS_CACHE):
        view = View.as_view()
        view(rf.post("/", {setting.setting: "New Value"}))


def test_analytics_settings_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:analytics:index"))
    assert response.status_code == 200


def test_captcha_settings_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:captcha:index"))
    assert response.status_code == 200


def test_general_settings_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:general:index"))
    assert response.status_code == 200


def test_oauth2_settings_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:oauth2:index"))
    assert response.status_code == 200


def test_threads_settings_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:threads:index"))
    assert response.status_code == 200


def test_users_settings_form_renders(admin_client):
    response = admin_client.get(reverse("misago:admin:settings:users:index"))
    assert response.status_code == 200



from . import store, gravatar, downloaded, dynamic, gallery, uploaded
from .default import set_default_avatar, set_default_avatar_from_url

__all__ = [
    "AVATAR_TYPES",
    "delete_avatar",
    "downloaded",
    "dynamic",
    "gallery",
    "gravatar",
    "set_default_avatar",
    "set_default_avatar_from_url",
    "store",
    "uploaded",
]

AVATAR_TYPES = ("gravatar", "dynamic", "gallery", "uploaded")

delete_avatar = store.delete_avatar
